"""Module interpret of viperleed.calc.files.parameters.

Initial version by @fkraushofer in 2020, major rewrite by @amimre
and @michele-riva in June 2023. This module used to be part of
parameters.py. Refactored in October 2023.

Functions and classes for interpreting the contents previously
read from a PARAMETERS file.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-18'
__license__ = 'GPLv3+'

from collections.abc import Sequence
import copy
from functools import partialmethod, wraps
import logging
from pathlib import Path
import re

import numpy as np
try:
    from matplotlib.colors import is_color_like  # For PLOT_IV
except ImportError:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True

from viperleed import __version__
from viperleed.calc.classes.rparams import EnergyRange
from viperleed.calc.classes.rparams import IVShiftRange
from viperleed.calc.classes.rparams import LayerCuts
from viperleed.calc.classes.rparams import SymmetryEps
from viperleed.calc.classes.rparams import TheoEnergies
from viperleed.calc.files.tenserleed import OLD_TL_VERSION_NAMES
from viperleed.calc.lib import periodic_table
from viperleed.calc.lib.base import parent_name
from viperleed.calc.lib.base import readIntRange, readVector
from viperleed.calc.lib.base import recombineListElements, splitSublists
from viperleed.calc.lib.version import Version
from viperleed.calc.lib.woods_notation import readWoodsNotation
from viperleed.calc.sections.calc_section import CalcSection as Section

from .checker import ParametersChecker
from .errors import ParameterBooleanConversionError
from .errors import ParameterConversionError
from .errors import ParameterError
from .errors import ParameterFloatConversionError
from .errors import ParameterHasNoValueError
from .errors import ParameterIntConversionError
from .errors import ParameterNeedsFlagError
from .errors import ParameterNeedsSlabError
from .errors import ParameterNotRecognizedError
from .errors import ParameterNumberOfInputsError
from .errors import ParameterParseError
from .errors import ParameterRangeError
from .errors import ParameterUnknownFlagError
from .errors import ParameterValueError
from .errors import SuperfluousParameterError
from .known_parameters import KNOWN_PARAMS, is_deprecated, warn_if_deprecated
from .utils import Assignment, NumericBounds, POSITIVE_FLOAT, POSITIVE_INT


_LOGGER = logging.getLogger(parent_name(__name__))


# Bool parameters for which to create interpret...() methods automatically.
# key is parameter name, value is tuple of keyword arguments that should be
# passed to interpret_bool_parameter, in order.
_SIMPLE_BOOL_PARAMS = {
    'LOG_SEARCH' : (),
    'PHASESHIFTS_CALC_OLD' : (),
    'PHASESHIFTS_OUT_OLD' : (),
    'R_FACTOR_LEGACY' : (),
    'STOP': (),
    'SUPPRESS_EXECUTION' : (),
    'SYMMETRIZE_INPUT' : (),
    'SYMMETRY_FIND_ORI' : (),
    'TL_IGNORE_CHECKSUM' : (),
    'LAYER_STACK_VERTICAL' : ({False: 'c', True: 'z'},),
    }


# Numerical parameters for which to create interpret...() methods
# automatically. Key is parameter name, value is a NumericBounds
_SIMPLE_NUMERICAL_PARAMS = {
    # Positive-only integers
    'BULKDOUBLING_MAX' : POSITIVE_INT,
    'N_CORES' : POSITIVE_INT,
    'SEARCH_MAX_GEN' : POSITIVE_INT,
    'TENSOR_INDEX' : POSITIVE_INT,
    # Positive-only floats
    'T_DEBYE' : POSITIVE_FLOAT,
    'T_EXPERIMENT' : POSITIVE_FLOAT,
    'V0_IMAG' : POSITIVE_FLOAT,
    # Other floats
    'V0_Z_ONSET' : NumericBounds(),
    'ATTENUATION_EPS' : NumericBounds(range_=(1e-6, 1),
                                      accept_limits=(True, False)),
    'BULKDOUBLING_EPS' : NumericBounds(range_=(1e-4, None),
                                       out_of_range_event='coerce'),
    'BULK_LIKE_BELOW': NumericBounds(range_=(0, 1),
                                     accept_limits=(False, False)),
    'SCREEN_APERTURE' : NumericBounds(range_=(0, 180)),
    # Other integers
    'HALTING' : NumericBounds(type_=int, range_=(1, 3)),
    'N_BULK_LAYERS' : NumericBounds(type_=int, range_=(1, 2)),
    'R_FACTOR_SMOOTH' : NumericBounds(type_=int, range_=(0, 999)),
    'R_FACTOR_TYPE' : NumericBounds(type_=int, range_=(1, 2)),
    'ZIP_COMPRESSION_LEVEL' : NumericBounds(type_=int, range_=(0, 9))
    }


# parameters that can be optimized in FD optimization
_OPTIMIZE_OPTIONS = {'theta', 'phi', 'v0i', 'a', 'b', 'c', 'ab', 'abc',}


def interpret(rpars, slab=None, silent=False):
    """Interpret rpars.readParams to actual values.

    Parameters
    ----------
    rpars : Rparams
        Object storing parameters for current run. Created previously
        by parameters.read, and should already contain raw string data.
    slab : Slab, optional
        Slab object with elements and atomic position data. If not
        passed, some parameters will not be interpreted.
    silent : bool, optional
        If True, less output will be printed. The default is False.

    Raises
    ------
    ParameterError
        If any parameter interpretation fails.
    """
    interpreter = ParameterInterpreter(rpars)
    interpreter.interpret(slab=slab, silent=silent)


def _with_dynamic_methods(cls):
    """Attach dynamically generated methods to the ParameterInterpreter cls."""

    def _interpret_deprecated(interpreter_method, param):
        """Interpret a deprecated PARAMETER."""
        @wraps(interpreter_method)
        def _wrapper_method(self, assignment, **kwargs):
            if warn_if_deprecated(param, __version__):
                self.rpars.setHaltingLevel(3)
            interpreter_method(self, assignment, **kwargs)
        return _wrapper_method

    # Dynamically produce methods for the 'simple parameters' listed
    # above. Notice that we should do this only once for the class,
    # and not for each instance separately. Calling the methods again
    # is possible, and would simply override the bindings
    cls.make_boolean_interpreter_methods()
    cls.make_numerical_interpreter_methods()

    # Now wrap all the interpreter methods for deprecated parameters
    for method_name, method in cls.__dict__.items():
        if not method_name.startswith('interpret_'):
            continue
        *_, param = method_name.split('interpret_')
        if not is_deprecated(param, __version__):
            continue
        setattr(cls, method_name, _interpret_deprecated(method, param))
    return cls


_BELOW_DEBUG = 2


# Disable 'too-many-public-methods' because of the mechanics of a
# ParameterInterpreter: a dispatcher for interpreting PARAMETERS.
# Methods are public, but are very rarely used. Most of the times
# they are used via the .interpret() interface.
@_with_dynamic_methods
class ParameterInterpreter:  # pylint: disable=too-many-public-methods
    """Class to interpret parameters from the PARAMETERS file.

    To add a new parameter, add a method with the name
    'interpret_<param>' and with a single assignment argument.
    """

    domains_ignore_params = {
        'BULK_REPEAT', 'ELEMENT_MIX', 'ELEMENT_RENAME',
        'LAYER_CUTS', 'N_BULK_LAYERS', 'SITE_DEF', 'SUPERLATTICE',
        'SYMMETRY_CELL_TRANSFORM', 'TENSOR_INDEX', 'TENSOR_OUTPUT'
        }

    grouplist = [                                                               # TODO: take from elsewhere
        'p1', 'p2', 'pm', 'pg', 'cm', 'rcm', 'pmm', 'pmg', 'pgg', 'cmm',
        'rcmm', 'p4', 'p4m', 'p4g', 'p3', 'p3m1', 'p31m', 'p6', 'p6m']

    bool_synonyms = {
        True: {'true', 'yes', '1', 't', 'y', 'on', 'enable', '.true.'},
        False: {'false', 'no', '0', 'f', 'n', 'off', 'disable', '.false.'}
        }

    def __init__(self, rpars):
        """Initialize interpreter instance from an Rparams."""
        self.rpars = rpars
        self.slab = None
        self._param_names = []  # In precedence order

        # Some flags
        self._search_conv_read = False

    def interpret(self, slab, silent=False):
        """Interpret all known parameters using slab."""
        self.slab = slab
        self._search_conv_read = False

        _backup_log_level = _LOGGER.level
        if silent:
            _LOGGER.setLevel(logging.ERROR)

        self._update_param_order()
        for param, assignment in self._get_param_assignments():
            self._complain_if_invalid_param_in_domain_calc(param)
            self._interpret_param(param, assignment)
            _LOGGER.log(_BELOW_DEBUG,
                        f'Successfully interpreted parameter {param}')

        # Make sure there is no inconsistent combination of parameters
        checker = ParametersChecker()
        checker.check_parameter_conflicts(self.rpars)

        # Finally set the log level back to what it was
        _LOGGER.setLevel(_backup_log_level)

    # ----------------  Helper methods for interpret() ----------------
    def _complain_if_invalid_param_in_domain_calc(self, param):
        """Raise if an invalid param is given in a DOMAINS calculation."""
        _is_domain_calc = 4 in self.rpars.RUN or self.rpars.domainParams
        if not _is_domain_calc or param not in self.domains_ignore_params:
            return
        err_ = (
            f'Found parameter {param} in the main PARAMETERS file of a '
            'DOMAIN calculation. This parameter can only be defined in '
            'the PARAMETERS files of the individual structural domains'
            )
        raise SuperfluousParameterError(param, message=err_)

    def _get_param_assignments(self):
        """Yield parameters and assignments for each PARAMETER read."""
        flat_params = (
            (param_name, assignment)
            for param_name in self._param_names
            for assignment in self.rpars.readParams[param_name]
            )
        yield from flat_params

    def _interpret_param(self, param, assignment):
        """Interpret the value of a single PARAMETER if known, or complain."""
        interpreter = getattr(self, f'interpret_{param.lower()}', None)
        if interpreter is None:
            # Complain about unknown parameter
            self.rpars.setHaltingLevel(2)
            raise ParameterNotRecognizedError(param)
        interpreter(assignment)

    def _update_param_order(self):
        """Define order in which parameters should be read."""
        ordered_params = 'LOG_LEVEL', 'RUN'
        self._param_names = [p for p in ordered_params
                             if p in self.rpars.readParams]
        self._param_names.extend(
            p for p in KNOWN_PARAMS
            if p in self.rpars.readParams and p not in self._param_names
            )

    # ---------- Methods for interpreting simple parameters -----------
    # Disable pylint warning since, while 6 is a bit many, we can't
    # really do much better than this. Merging some in a container
    # does not seem clearer.
    # pylint: disable-next=too-many-arguments
    def interpret_bool_parameter(self, assignment,
                                 allowed_values=None,
                                 param=None, return_only=False,
                                 no_flags=False):
        """Set a parameter to a boolean value.

        Parameters
        ----------
        assignment : Assignment
            The assignment of the parameter, read from PARAMETERS.
        allowed_values : dict, optional
            Additional string values which should be interpreted as
            False or True. E.g. by default, 'f' and 'false' are False,
            't' and 'true' are True. Keys should be True and False,
            values are Sequences of acceptable strings.
        param : str, optional
            The name of the parameter in PARAMETERS, if it differs
            from assignment.parameter. Also used as attribute name
            for self.rpars. Default is None.
        return_only: bool, optional
            If True, only return the value of the parameter, but do
            not set it. Default is False.
        no_flags : bool, optional
            If True, complain if assignment has flags. Default is
            False.

        Returns
        -------
        value : bool
            The value of the parameter.

        Raises
        ------
        ValueError
            If allowed_values contains keys other than True and False.
        ValueError
            If allowed_values contains identical aliases for True
            and False.
        ParameterUnknownFlagError
            When assignment has flags but no_flag is True.
        ParameterBooleanConversionError
            If the string value of the parameter does not
            have an acceptable boolean correspondent.
        """
        if allowed_values is None:
            allowed_values = {}
        if no_flags:
            self._ensure_no_flags_assignment(assignment, param)

        _bool_synonyms = copy.deepcopy(self.bool_synonyms)
        for option, values in allowed_values.items():
            try:
                _bool_synonyms[option].update(v.lower() for v in values)
            except KeyError:
                raise ValueError(f'Unexpected option={option} '
                                 'in allowed_values') from None

        # Make sure there is no intersection between the two sets
        if set(_bool_synonyms[True]) & set(_bool_synonyms[False]):
            raise ValueError('The sets of allowed values for '
                             'True and False must not overlap')

        # Check if the value is in the allowed ones
        str_value = assignment.value.lower()
        try:
            value = next(bool_ for bool_, synonyms in _bool_synonyms.items()
                         if str_value in synonyms)
        except StopIteration:  # Value is invalid
            self.rpars.setHaltingLevel(1)
            raise ParameterBooleanConversionError(assignment.parameter,
                                                  assignment.value) from None

        param = param or assignment.parameter
        if not return_only:
            setattr(self.rpars, param.upper(), value)
        return value

    # Disable pylint warning since, while 6 is a bit many, we can't
    # really do much better than this. Merging some in a container
    # does not seem clearer.
    # pylint: disable-next=too-many-arguments
    def interpret_numerical_parameter(self, assignment,
                                      param=None, return_only=False,
                                      bounds=NumericBounds(),                   # TODO: ideally one could default to actually using the limits known from self.rpars.get_limits!
                                      no_flags=False):
        """Set a parameter to a numeric (int or float) value.

        Parameters
        ----------
        assignment : Assignment
            The assignment of the parameter, read from PARAMETERS.
        param : str, optional
            The name of the parameter in PARAMETERS, if it differs
            from assignment.parameter. Also used as attribute name
            for self.rpars. Default is None.
        return_only: bool, optional
            If True, only return the value of the parameter, but
            do not set it. Default is False.
        bounds : NumericBounds, optional
            Acceptable limits for value, and what to do in case an
            out-of-bounds value is given. Default is an unbounded
            float.
        no_flags : bool, optional
            If True, complain if assignment has flags. Default is
            False.

        Returns
        ----------
        value : int or float
            The interpreted value.

        Raises
        ------
        ParameterFloatConversionError
            When conversion of a float-type value to float fails.
        ParameterIntConversionError
            When conversion of a int-type value to int fails.
        ParameterRangeError
            When a the value falls outside the constraints of bounds.
        ParameterUnknownFlagError
            When assignment has flags but no_flag is True.
        """
        if no_flags:
            self._ensure_no_flags_assignment(assignment, param)
        type_ = bounds.type_
        exc_cls = (ParameterFloatConversionError if type_ is float
                   else ParameterIntConversionError)
        exc = exc_cls(assignment.parameter, assignment.value)

        try:
            # First convert to float. Necessary for, e.g., 1e6 as int
            float_value = float(assignment.value)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise exc from None

        value = type_(float_value)
        if type_ is int and not np.isclose(value, float_value):
            raise exc

        in_range = all(bounds.is_in_range(value))
        if not in_range and bounds.fail:
            self.rpars.setHaltingLevel(1)
            out_of_range = bounds.format_out_of_range(value)
            raise ParameterRangeError(assignment.parameter,
                                      message=out_of_range)

        if not in_range:
            in_range_value = bounds.make_in_range(value)
            out_of_range = bounds.format_out_of_range(value)
            _LOGGER.warning(f'PARAMETERS file: {assignment.parameter}: '
                            f'{out_of_range}. '
                            f'Value will be set to {in_range_value}.')
            value = in_range_value

        param = param or assignment.parameter
        if not return_only:
            setattr(self.rpars, param.upper(), value)
        return value

    @classmethod
    def _make_methods(cls, wrapped, kwargs_names, new_methods_info):
        """Dynamically add methods for this class.

        Parameters
        ----------
        wrapped : callable
            The callable that will be wrapped to create methods.
            The call to new_method(*args, **kwargs) becomes a
            wrapped(*args, **wrapped_kwargs, **kwargs) call,
            where wrapped_kwargs is generated here using
            kwargs_names and new_methods_info.
        kwargs_names : Sequence
            Names (str) of the keyword arguments that may be
            replaced. At most len(kwargs_names) keyword arguments
            will be given to wrapped.
        new_methods_info : dict
            Keys are names of the parameters for which methods are to
            be generated. They are used as the the first positional
            argument to wrapped. Also, the new methods will be
            named "interpret_<key>". Values are sequences of the
            values of keyword arguments to be passed on to wrapped,
            in the same order.

        Raises
        ------
        TypeError
            If too many keyword argument values are given in any of
            the new_methods_info
        """
        for param, kwargs_values in new_methods_info.items():
            method_name = f'interpret_{param.lower()}'
            if not isinstance(kwargs_values, Sequence):
                kwargs_values = (kwargs_values,)
            if len(kwargs_values) > len(kwargs_names):
                raise TypeError(
                    f'Too many keyword arguments for {method_name}. '
                    f'Expected at most values for {kwargs_names}'
                    )
            kwargs = dict(zip(kwargs_names, kwargs_values))
            kwargs['param'] = param
            kwargs['no_flags'] = True  # no flags for simple parameters
            method = partialmethod(wrapped, **kwargs)
            setattr(cls, method_name, method)

    @classmethod
    def make_boolean_interpreter_methods(cls):
        """Dynamically generate bool-setting methods."""
        cls._make_methods(cls.interpret_bool_parameter,
                          ('allowed_values',),
                          _SIMPLE_BOOL_PARAMS)

    @classmethod
    def make_numerical_interpreter_methods(cls):
        """Dynamically generate int/float-setting methods."""
        cls._make_methods(cls.interpret_numerical_parameter,
                          ('bounds',),
                          _SIMPLE_NUMERICAL_PARAMS)

    # ----- Methods to make sure no extra flags/values are given ------
    def _ensure_single_flag_assignment(self, assignment, param=None,
                                       message=None,
                                       must_have_exaclty_one=True):
        """Raise if assignment contains multiple or, optionally, no flags."""
        if param is None:
            param = assignment.parameter
        if must_have_exaclty_one and not assignment.flag:
            self.rpars.setHaltingLevel(1)
            raise ParameterNeedsFlagError(param, message)
        if message is None:
            message = assignment.flags_str
        if assignment.other_flags:
            self.rpars.setHaltingLevel(1)
            raise ParameterUnknownFlagError(param, message)

    def _ensure_single_value_assignment(self, assignment, param=None):
        """Raise if assignment contains multiple values."""
        n_values = len(assignment.values)
        if param is None:
            param = assignment.parameter
        if not n_values:
            self.rpars.setHaltingLevel(1)
            raise ParameterHasNoValueError(param)
        if n_values != 1:
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param, (n_values, 1))

    def _ensure_no_flags_assignment(self, assignment, param=None):
        """Raise if assignment contains any flag."""
        if param is None:
            param = assignment.parameter
        if assignment.flags:
            self.rpars.setHaltingLevel(1)
            raise ParameterUnknownFlagError(param, assignment.flag)

    def _ensure_simple_assignment(self, assignment, param=None):
        """Raise if assignment is not simple (i.e., one value, no flags)."""
        self._ensure_no_flags_assignment(assignment, param)
        self._ensure_single_value_assignment(assignment, param)

    def _ensure_single_flag_and_value_assignment(self, assignment, param=None):
        """Raise if assignment does not have a single flag--value pair."""
        self._ensure_single_flag_assignment(assignment, param)
        self._ensure_single_value_assignment(assignment, param)

    # ----------- Methods to interpret individual parameters ----------
    def interpret_average_beams(self, assignment):                              # TODO: Issue #135
        """Assign parameter AVERAGE_BEAMS."""
        param = 'AVERAGE_BEAMS'
        self._ensure_no_flags_assignment(assignment)
        right_side = assignment.values_str.lower().strip()
        # trivial cases
        if right_side in {'off', 'none', 'false', 'f'}:
            self.rpars.AVERAGE_BEAMS = False
            return
        if right_side.lower() in {'all', 'perpendicular', 'perp'}:
            self.rpars.AVERAGE_BEAMS = (0., 0.)
            return

        # Otherwise, try to parse the value
        angles = self._parse_incidence_angles(param, assignment)
        self.rpars.AVERAGE_BEAMS = angles['THETA'], angles['PHI']

    def interpret_beam_incidence(self, assignment):
        """Assign parameter BEAM_INCIDENCE."""
        param = 'BEAM_INCIDENCE'
        self._ensure_no_flags_assignment(assignment)
        angles = self._parse_incidence_angles(param, assignment)
        self.rpars.THETA, self.rpars.PHI = angles['THETA'], angles['PHI']

    def interpret_bulk_repeat(self, assignment):
        """Assign parameter BULK_REPEAT."""
        param = 'BULK_REPEAT'
        if not self.slab:  # BULK_REPEAT is moot without a slab
            raise ParameterNeedsSlabError(param)

        bulk_repeat_str = assignment.values_str.lower()
        # (1) Vector
        if '[' in bulk_repeat_str:
            vec = readVector(bulk_repeat_str, self.slab.ucell)
            if vec is None:
                raise ParameterParseError(parameter=param)
            self.rpars.BULK_REPEAT = vec
            return

        # (2) Z distance
        if '(' not in bulk_repeat_str:
            try:
                self.rpars.BULK_REPEAT = abs(float(assignment.value))
            except ValueError:
                raise ParameterFloatConversionError(parameter=param) from None
            return

        # (3) C or Z distance. Should match, e.g., c(2.0) or z(2.0)
        match = re.match(r'\s*(c|z)\(\s*(?P<val>[0-9.]+)\s*\)',
                         bulk_repeat_str)
        if not match:
            raise ParameterParseError(parameter=param)
        try:
            val = abs(float(match.group('val')))
        except ValueError:
            raise ParameterFloatConversionError(parameter=param) from None

        val = self.slab.c_vector[2] * val if 'c' in bulk_repeat_str else val
        self.rpars.BULK_REPEAT = val

    def interpret_domain(self, assignment):
        """Set the domain path and name."""
        param = 'DOMAIN'
        # Check name
        name = assignment.flag
        if name in self.rpars.DOMAINS:  # Already defined
            error_message = f'Multiple sources defined for domain {name}'
            self.rpars.setHaltingLevel(3)
            raise ParameterValueError(param, message=error_message)
        if not name:  # Get unique name                                         # TODO: used in several other places
            i = 1
            while str(i) in self.rpars.DOMAINS:
                i += 1
            name = str(i)

        # Check path
        right_side = assignment.values_str.strip()
        if Path(right_side).exists():
            path = right_side
        elif Path(right_side).with_suffix('.zip').is_file():
            path = right_side + '.zip'
        else:
            error_message = (f'Value for DOMAIN {name} could not be '
                             'interpreted as either a path or a .zip file')
            self.rpars.setHaltingLevel(3)
            raise ParameterValueError(param, message=error_message)
        self.rpars.DOMAINS[name] = path

    def interpret_domain_step(self, assignment):
        """Assign parameter DOMAIN_STEP."""
        param = 'DOMAIN_STEP'
        self._ensure_simple_assignment(assignment)
        domain_step = self.interpret_numerical_parameter(
            assignment,
            bounds=NumericBounds(type_=int, range_=(1, 100)),
            return_only=True
            )

        # pylint: disable=compare-to-zero
        # Seems clearer this way than having "if 100 % domain_step"
        if 100 % domain_step != 0:
            j = domain_step - 1
            while 100 % j != 0:
                j -= 1
            message = (f'100 is not divisible by given value {domain_step}. '
                       f'Consider using {j} instead')
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, message=message)
        self.rpars.DOMAIN_STEP = domain_step

    def interpret_element_mix(self, assignment):
        """Assign parameter ELEMENT_MIX."""
        param = 'ELEMENT_MIX'
        self._ensure_single_flag_assignment(assignment)
        element = self._get_valid_slab_element_from_flag(param, assignment)

        self._ensure_chemical_elements(param, assignment.values)
        mix = [el.capitalize() for el in assignment.values]
        if len(mix) == 1:
            rename = mix[0]
            self.rpars.setHaltingLevel(3)
            message = (f'Only one mixed element found: {rename}. If you '
                       f'intended to rename {element} to {rename}, you should '
                       f'use "ELEMENT_RENAME {element} = {rename}"')
            raise ParameterNumberOfInputsError(param, message=message)
        self.rpars.ELEMENT_MIX[element] = mix

    def interpret_element_rename(self, assignment):
        """Assign parameter ELEMENT_RENAME."""
        param = 'ELEMENT_RENAME'
        self._ensure_single_flag_and_value_assignment(assignment)
        element = self._get_valid_slab_element_from_flag(param, assignment)
        self._ensure_chemical_elements(param, assignment.values)
        self.rpars.ELEMENT_RENAME[element] = assignment.value.capitalize()

    def interpret_filament_wf(self, assignment):
        """Assign parameter FILAMENT_WF."""
        param = 'FILAMENT_WF'
        self._ensure_simple_assignment(assignment)
        # Check common filaments (e.g., W), otherwise assume a float
        known_filaments = self.rpars.get_default(param)
        try:
            self.rpars.FILAMENT_WF = known_filaments[assignment.value.lower()]
        except KeyError:
            self.interpret_numerical_parameter(assignment)

    def interpret_fortran_comp(self, assignment):                               # TODO: Issue #136
        """Assign parameter FORTRAN_COMP."""
        param = 'FORTRAN_COMP'
        message = (f'Only one flag allowed for {param} per line. '
                   f'Got {assignment.flags}')
        self._ensure_single_flag_assignment(assignment, message=message,
                                            must_have_exaclty_one=False)
        flag, compiler_str = assignment.flag.lower(), assignment.values_str

        # (1) Default (i.e., non-MPI) compiler flags
        if not flag and compiler_str.lower() in {'ifort', 'gfortran'}:
            self.rpars.FORTRAN_COMP[0] = compiler_str.lower()
            return

        # (2) MPI compiler flags
        if flag == 'mpi' and compiler_str.lower() in {'mpifort', 'mpiifort'}:
            self.rpars.FORTRAN_COMP_MPI[0] = compiler_str.lower()
            return

        # (3) Custom compiler flags or full compilation string.
        #     Remove optional quotes
        delim = assignment.values_str[0]
        if delim in {'"', "'"}:
            compiler_str = assignment.values_str[1:]
            if not compiler_str.endswith(delim):
                raise ParameterValueError(param,
                                          message='No closing quotation mark')
            compiler_str = compiler_str[:-1]

        if not flag:
            self.rpars.FORTRAN_COMP[0] = compiler_str
        elif flag == 'post':
            self.rpars.FORTRAN_COMP[1] = compiler_str
        elif flag == 'mpi':
            self.rpars.FORTRAN_COMP_MPI[0] = compiler_str
        elif flag == 'mpipost':
            self.rpars.FORTRAN_COMP_MPI[1] = compiler_str
        else:
            raise ParameterUnknownFlagError(parameter=param,
                                            flag=assignment.flag)

    def interpret_intpol_deg(self, assignment):
        """Assign parameter INTPOL_DEG."""
        param = 'INTPOL_DEG'
        self._ensure_simple_assignment(assignment)
        intpol_deg = assignment.value
        if intpol_deg in self.rpars.get_limits(param):
            self.rpars.INTPOL_DEG = int(intpol_deg)
            return
        self.rpars.setHaltingLevel(1)
        message = 'Only degree 3 and 5 interpolation supported at the moment'
        raise ParameterValueError(param, message=message)

    def interpret_iv_shift_range(self, assignment):
        """Assign parameter IV_SHIFT_RANGE."""
        param = 'IV_SHIFT_RANGE'
        energies = self._parse_energy_range(assignment, assignment.values)
        if len(energies) not in (2, 3):
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(parameter=param)

        try:
            iv_range = IVShiftRange(*energies)
        except (ValueError, RuntimeError) as exc:
            # RuntimeError: was fixed, and we tried to unfix it
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, message=str(exc)) from exc

        # Make sure that the user gave us a reasonable
        # step. Inform them if we had to change stuff
        if iv_range.defined and not iv_range.is_equivalent(energies):
            _LOGGER.info(f'IV_SHIFT_RANGE bounds were not integer multiples '
                         f'of {iv_range.step}. Bounds have been adjusted to '
                         f'{iv_range.min} {iv_range.max}.')

        # Interpret underscores as defaults
        iv_range.set_undefined_values(self.rpars.get_default(param))
        self.rpars.IV_SHIFT_RANGE = iv_range

    def interpret_layer_cuts(self, assignment):
        """Assign parameter LAYER_CUTS."""
        param = 'LAYER_CUTS'
        try:
            cuts = LayerCuts.from_string(assignment.values_str)
        except ValueError as exc:
            # String has invalid syntax
            raise ParameterParseError(param, message=str(exc)) from None
        if not cuts:
            raise ParameterValueError(param,
                                      message='At least one layer cut needed')
        self.rpars.LAYER_CUTS = cuts

    @staticmethod
    def _tokenize_int_range(numeric_string):
        r"""Return integers representing range bounds in `numeric_string`.

        A token of the form '-\d+' is considered as a negative number
        unless it is preceded by another '\d+' without delimiters, or
        only by '-' delimiters. This means that
        >>> _tokenize_int_range('1-15')
        [1, 15]
        >>> _tokenize_int_range('-1---15')
        [-1, 15]
        >>> _tokenize_int_range('-1-- -15')
        [-1, -15]
        >>> _tokenize_int_range('1 -15')
        [1, -15]

        Acceptable delimiters are spaces, '-', or ':'.

        Parameters
        ----------
        numeric_string : str
            The string to be analysed.

        Returns
        -------
        integers : list
            Integers in `numeric_string`.

        Raises
        ------
        ValueError
            If any of the entries is neither a number or a delimiter
        ValueError
            If, after stripping spaces, `numeric_string` ends with a
            delimiter instead of a number
        """
        string = re.sub(r'\s', ' ', numeric_string.strip())
        delimiter_chars = ' :-'
        delimiters = tuple(delimiter_chars)
        # Remove all adjacent identical delimiters
        for delimiter in delimiter_chars:
            string = re.sub(fr'{delimiter}*{delimiter}', delimiter, string)
        if not string:
            return []
        _digits_or_delim_re = re.compile(fr'[\d{delimiter_chars}]+')
        _digits_re = re.compile(r'\d+')
        if not _digits_or_delim_re.fullmatch(string):
            raise ValueError('contains non-integers or invalid delimiters')
        if string.endswith(delimiters):
            raise ValueError('ends with a delimiter')
        integers = []
        for match_digits in _digits_re.finditer(string):
            token, start = match_digits.group(), match_digits.start()
            delimiter = string[start-1:start]
            if not delimiter:  # First token in string
                integers.append(int(token))
                continue
            if delimiter != '-':
                integers.append(int(token))
                continue
            # Figure out if the '-' is a delimiter or it means negative
            previous = string[:max(start-1, 0)]
            if not previous or previous.endswith(delimiters):
                # Negative number
                integers.append(int(delimiter + token))
            else:
                integers.append(int(token))
        return integers

    def interpret_lmax(self, assignment):
        """Assign parameter LMAX."""
        param = 'LMAX'
        try:
            values = self._tokenize_int_range(assignment.values_str)
        except ValueError as exc:
            raise ParameterValueError(
                param,
                message=f'Invalid {assignment.values_str!r}: {exc}'
                ) from None

        if not values:
            self.rpars.setHaltingLevel(1)
            raise ParameterHasNoValueError(param)
        if len(values) > 2:
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param)
        try:
            self.rpars.LMAX = self.rpars.LMAX.from_value(values)
        except ValueError as exc:  # out-of-range
            raise ParameterRangeError(param, message=str(exc))

    def interpret_log_level(self, assignment):
        """Assign parameter LOG_LEVEL."""
        param = 'LOG_LEVEL'
        self._ensure_simple_assignment(assignment)

        # Try to interpret as bool. This is the way
        # the previous LOG_DEBUG used to work
        try:
            log_debug = self.interpret_bool_parameter(assignment,
                                                      return_only=True)
        except ParameterError:
            pass
        else:
            self.rpars.LOG_LEVEL = logging.DEBUG if log_debug else logging.INFO
            return

        # Try to interpret as string (e.g. 'verbose', 'vv')
        _defaults = self.rpars.get_default(param)
        if assignment.values_str.lower() in _defaults:
            self.rpars.LOG_LEVEL = _defaults[assignment.values_str.lower()]
            return

        # Otherwise interpret as int
        try:
            log_level = self.interpret_numerical_parameter(
                assignment,
                bounds=NumericBounds(type_=int, range_=(0, 50)),
                return_only=True
                )
        except ParameterError as exc:
            raise ParameterValueError(param, assignment.values_str) from exc
        self.rpars.LOG_LEVEL = log_level

    def interpret_optimize(self, assignment):
        """Assign parameter OPTIMIZE."""
        param = 'OPTIMIZE'
        if Section.FD_OPTIMIZATION.value not in self.rpars.RUN:                 # TODO: remove .value when using CalcSection in RUN
            msg_ = ('RUN does not include a full-dynamic-optimization section '
                    f'(RUN = {Section.FD_OPTIMIZATION.value}). It makes no '
                    'sense to provide an OPTIMIZE parameter. OPTIMIZE will be '
                    'ignored.')
            _LOGGER.warning(msg_)
            self.rpars.setHaltingLevel(1)
            return
        if not assignment.flag:
            message = 'Parameter to optimize not defined'
            self.rpars.setHaltingLevel(3)
            raise ParameterNeedsFlagError(param, message)
        which = assignment.flag.lower()
        if which not in _OPTIMIZE_OPTIONS:
            self.rpars.setHaltingLevel(3)
            raise ParameterUnknownFlagError(param, f'{which!r}')
        self.rpars.OPTIMIZE['which'] = which
        if not assignment.other_values:
            try:
                self.rpars.OPTIMIZE['step'] = float(assignment.value)
            except ValueError:
                self.rpars.setHaltingLevel(1)
                raise ParameterFloatConversionError(param,
                                                    assignment.value) from None
            return
        for flag_value_pair in assignment.values_str.split(','):
            self._interpret_optimize_flag_value_pair(param, flag_value_pair)

    def _interpret_optimize_flag_value_pair(self, param, flag_value_pair):
        """Interpret one 'flag value' pair for OPTIMIZE."""
        flag, value = self._get_flag_value_from_pair(param, flag_value_pair)
        value_error = f'Value {value!r} is invalid for flag {flag!r}'
        partype = {'step': float, 'convergence': float,
                   'minpoints': int, 'maxpoints': int,
                   'maxstep': float}
        try:
            numeric = partype[flag](value)
        except ValueError as exc:
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, message=value_error) from exc
        except KeyError:
            self.rpars.setHaltingLevel(2)
            raise ParameterUnknownFlagError(param, f'{flag!r}') from None
        self.rpars.OPTIMIZE[flag] = numeric

    def interpret_parabola_fit(self, assignment):                               # TODO: Issue #137
        """Assign parameter PARABOLA_FIT."""
        if assignment.values_str.lower() == 'off':
            self.rpars.PARABOLA_FIT['type'] = 'none'
            return
        for flag_value_pair in assignment.values_str.split(','):
            self._interpret_parabola_fit_flag_value_pair(flag_value_pair)

    def _interpret_parabola_fit_flag_value_pair(self, flag_value_pair):
        """Interpret one 'flag value' pair for PARABOLA_FIT."""
        param = 'PARABOLA_FIT'
        flag, value = self._get_flag_value_from_pair(param, flag_value_pair)
        if flag == 'localise':
            flag = 'localize'
        value_error = (f'Invalid value {value} for flag {flag}. '
                       'Value will be ignored')
        if flag not in ('type', 'alpha', 'mincurv', 'localize'):
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, f'Unknown flag={flag!r}')
        if flag == 'type' and value not in ('linear', 'linearregression',
                                            'lasso', 'ridge', 'elasticnet',
                                            'none'):
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, value_error)
        if flag == 'type':
            self.rpars.PARABOLA_FIT[flag] = value
            return
        try:
            value_float = float(value)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise ParameterFloatConversionError(param, value) from None
        if value_float >= 0:
            self.rpars.PARABOLA_FIT[flag] = value_float
            return
        self.rpars.setHaltingLevel(1)
        raise ParameterRangeError(param,
                                  message=f'{flag} value must be non-negative')

    def interpret_phaseshift_eps(self, assignment):
        """Assign parameter PHASESHIFT_EPS."""
        param = 'PHASESHIFT_EPS'
        self._ensure_simple_assignment(assignment)
        try:
            ps_eps = float(assignment.value)
        except ValueError:
            # check if one of default values (e.g. 'fine')
            preset = assignment.value.lower()[0]
            _defaults = self.rpars.get_default(param)
            ps_eps = _defaults.get(preset, None)
            if ps_eps is None:
                self.rpars.setHaltingLevel(1)
                raise ParameterFloatConversionError(param) from None
        if 0 < ps_eps < 1:
            self.rpars.PHASESHIFT_EPS = ps_eps
        else:
            self.rpars.setHaltingLevel(1)
            raise ParameterRangeError(param, given_value=ps_eps,
                                      allowed_range=(0, 1))

    def interpret_plot_iv(self, assignment):
        """Assign parameter PLOT_IV."""
        if not _CAN_PLOT:
            # Cannot interpret this parameter
            return
        param = 'PLOT_IV'
        self._ensure_single_flag_assignment(assignment)
        flag_aliases = {
            'plot': 'plot',
            'axes': 'axes', 'border': 'axes', 'borders': 'axes',
            'colors': 'colors', 'colours': 'colors',
            'color': 'colors', 'colour': 'colors',
            'font': 'font_size', 'fontsize': 'font_size',
            'font_size': 'font_size',
            'legend': 'legend', 'legends': 'legend',
            'lw': 'line_width', 'linewidth': 'line_width',
            'line_width': 'line_width', 'line': 'line_width',
            'overbar': 'overbar', 'overline': 'overbar',
            'perpage': 'perpage', 'layout': 'perpage',
            }
        flag = assignment.flag.lower()
        try:
            flag = flag_aliases[flag]
        except KeyError:
            self.rpars.setHaltingLevel(1)
            raise ParameterUnknownFlagError(param, f'{flag!r}') from None
        setter = getattr(self, f'_interpret_plot_iv__{flag}')
        setter(assignment)

    def _interpret_plot_iv__axes(self, assignment):
        """Assign PLOT_IV['axes']."""
        self._ensure_single_value_assignment(assignment)
        synonyms = {'all': 'all',
                    'none': 'none',
                    'lb': 'lb', 'less': 'lb',
                    'b': 'b', 'bottom': 'b',}
        try:
            self.rpars.PLOT_IV['axes'] = synonyms[assignment.value.lower()]
        except KeyError:
            self.rpars.setHaltingLevel(1)
            raise ParameterParseError(assignment.parameter) from None

    def _interpret_plot_iv__colors(self, assignment):
        """Assign PLOT_IV['colors']."""
        colors = assignment.values
        if not colors:
            self.rpars.setHaltingLevel(1)
            raise ParameterHasNoValueError(assignment.parameter)
        if not all(is_color_like(c) for c in colors):
            err_ = ('Cannot interpret at least one of '
                    f'{assignment.values_str!r} as color(s)')
            raise ParameterValueError(assignment.parameter, message=err_)
        self.rpars.PLOT_IV['colors'] = colors

    def _interpret_plot_iv__font_size(self, assignment):
        """Assign PLOT_IV['font_size']."""
        self._ensure_single_value_assignment(assignment)
        try:
            font_size = float(assignment.value)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise ParameterFloatConversionError(assignment.parameter,
                                                assignment.value) from None
        if font_size <= 0:
            self.rpars.setHaltingLevel(1)
            raise ParameterRangeError(assignment.parameter,
                                      message='Font size must be positive.')
        self.rpars.PLOT_IV['font_size'] = font_size

    def _interpret_plot_iv__legend(self, assignment):
        """Assign PLOT_IV['legend']."""
        self._ensure_single_value_assignment(assignment)
        value = assignment.value.lower()
        if value in {'all', 'first', 'none'}:
            self.rpars.PLOT_IV['legend'] = value
        elif value in {'topright', 'tr'}:
            self.rpars.PLOT_IV['legend'] = 'tr'
        else:
            self.rpars.setHaltingLevel(1)
            raise ParameterParseError(assignment.parameter)

    def _interpret_plot_iv__line_width(self, assignment):
        """Assign PLOT_IV['line_width']."""
        self._ensure_single_value_assignment(assignment)
        try:
            line_width = float(assignment.value)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise ParameterFloatConversionError(assignment.parameter,
                                               assignment.value) from None
        if line_width <= 0:
            self.rpars.setHaltingLevel(1)
            raise ParameterRangeError(assignment.parameter,
                                      message='Line width must be > 0.')
        self.rpars.PLOT_IV['line_width'] = line_width

    def _interpret_plot_iv__overbar(self, assignment):
        """Assign PLOT_IV['overbar']."""
        self._ensure_single_value_assignment(assignment)
        value = assignment.value.lower()
        if value.startswith('t'):
            self.rpars.PLOT_IV['overbar'] = True
        elif value.startswith('f'):
            self.rpars.PLOT_IV['overbar'] = False
        else:
            message = f'Value for flag {assignment.flag!r} not recognized'
            self.rpars.setHaltingLevel(1)
            raise ParameterParseError(assignment.parameter, message)

    def _interpret_plot_iv__perpage(self, assignment):
        """Assign PLOT_IV['perpage']."""
        # Can be one or two integers
        if not assignment.values:
            self.rpars.setHaltingLevel(1)
            raise ParameterHasNoValueError(assignment.parameter)
        if len(assignment.values) > 2:
            self.rpars.setHaltingLevel(1)
            message = 'perpage accepts one or two positive integers'
            raise ParameterNumberOfInputsError(assignment.parameter,
                                               message=message)
        try:
            values = tuple(int(v) for v in assignment.values)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise ParameterIntConversionError(assignment.parameter,
                                              assignment.values) from None
        if any(v <= 0 for v in values):
            _plural = 's' if len(values)>1 else ''
            message = f'perpage value{_plural} must be positive'
            raise ParameterRangeError(assignment.parameter, message=message)
        self.rpars.PLOT_IV['perpage'] = values if len(values)>1 else values[0]

    def _interpret_plot_iv__plot(self, assignment):
        """Assign PLOT_IV['plot']."""
        self._ensure_single_value_assignment(assignment)
        extra_synonyms = {False: ('none',)}
        self.rpars.PLOT_IV['plot'] = self.interpret_bool_parameter(
            assignment, extra_synonyms, return_only=True
            )

    def interpret_run(self, assignment):
        """Assign parameter RUN, inserting an initialization if needed."""
        param = 'RUN'
        segments = []
        for section_str in assignment.values:
            try:
                segments.extend(Section.sequence_from_string(section_str))
            except ValueError as exc:
                self.rpars.setHaltingLevel(2)
                raise ParameterValueError(param, section_str) from exc
        if Section.DOMAINS in segments:
            _LOGGER.info('Found domain search.')
        if not segments:
            self.rpars.setHaltingLevel(3)
            message = f'{param} was defined, but no values were read'
            raise ParameterHasNoValueError(param, message)
        # Insert initialization section if not present
        if segments[0] is not Section.INITIALIZATION:
            segments.insert(0, Section.INITIALIZATION)
        self.rpars.RUN = [s.value for s in segments]                            # TODO: replace with "segments" to keep Section objects

    def interpret_search_beams(self, assignment):
        """Assign parameter SEARCH_BEAMS."""
        param = 'SEARCH_BEAMS'
        self._ensure_simple_assignment(assignment)
        value = assignment.value.lower()
        if value.startswith(('0', 'a')):
            self.rpars.SEARCH_BEAMS = 0
        elif value.startswith(('1', 'i')):
            self.rpars.SEARCH_BEAMS = 1
        elif value.startswith(('2', 'f')):
            self.rpars.SEARCH_BEAMS = 2
        else:
            raise ParameterValueError(param, value)

    def interpret_search_convergence(self, assignment, is_updating=False):      # TODO: Issue #139
        """Interpret SEARCH_CONVERGENCE parameter.

        This method also sets rpars.searchConvInit if `is_updating`
        is True-thy and the values to be assigned have changed.

        Parameters
        ----------
        assignment : Assignment
            The assignment line, containing details of the
            flags and values to be interpreted.
        is_updating : bool, optional
            Whether this method is being called as part of a
            parameters.update (True) or a parameters.interpret
            (False). Default is False.

        Raises
        ------
        ParameterHasNoValueError
            If assignment carries no values.
        ParameterNumberOfInputsError
            If there are more than two values for this parameter.
        ParameterNeedsFlagError
            If more values are present, but no flag was given.
        ParameterUnknownFlagError
            If the first flag is neither 'dgen' nor 'gaussian'.
        ParameterUnknownFlagError
            If the first flag is 'gaussian' and there are more
            flags.
        ParameterUnknownFlagError
            If first flag is 'dgen' and there is more than one
            additional flag, or if the additional flag is not
            one of the acceptable ones ('dec', 'best', 'all').
        ParameterFloatConversionError
            If numeric values cannot be converted to float.
        ParameterRangeError
            If Gaussian width, its scaling factor, dgen, or its
            scaling factor are out of bounds
        """
        param = 'SEARCH_CONVERGENCE'
        if not assignment.values:
            raise ParameterHasNoValueError(param)
        if len(assignment.values) > 2:
            raise ParameterNumberOfInputsError(param)

        if (not assignment.flags
                and not assignment.other_values
                and assignment.value.lower() == 'off'):
            self.rpars.GAUSSIAN_WIDTH_SCALING = 1.
            return

        if not assignment.flags:
            raise ParameterNeedsFlagError(param)
        if assignment.flag.lower() not in ('dgen', 'gaussian'):
            raise ParameterUnknownFlagError(param, f'{assignment.flag!r}')

        numeric = [None, None]
        for i, value in enumerate(assignment.values):
            try:
                numeric[i] = float(value)
            except ValueError as err:
                raise ParameterFloatConversionError(param) from err

        if assignment.flag.lower() == 'gaussian':
            self._ensure_single_flag_assignment(assignment)
            self._interpret_search_convergence_gaussian(
                param, *numeric, is_updating
                )
        else:
            self._interpret_search_convergence_dgen(
                param, assignment.other_flags, *numeric, is_updating
                )

    # disable: The only option would be to pack max_dgen and scaling
    # into a single object, but I'm not sure this would make it
    # more understandable.
    # pylint: disable-next=too-many-arguments
    def _interpret_search_convergence_dgen(self, param, flags, max_dgen,
                                           scaling, is_updating):
        """Assign the SEARCH_MAX_DGEN/_SCLING parameters."""
        if len(flags) > 1:
            raise ParameterUnknownFlagError(param, ' '.join(flags[1:]))
        target = flags[0] if flags else 'dec'
        if target not in ['dec', 'best', 'all']:
            raise ParameterUnknownFlagError(param, f'{target!r}')
        should_store_new_value = (
            not is_updating
            or max_dgen != self.rpars.searchConvInit['dgen'][target]
            )
        valid = max_dgen is not None and max_dgen > 0
        if should_store_new_value and not valid:
            message = 'dgen should be a positive number'
            raise ParameterRangeError(param, message=message)

        if should_store_new_value and not self._search_conv_read:
            # Clear default values
            self.rpars.reset_default('SEARCH_MAX_DGEN')
            self._search_conv_read = True
        if should_store_new_value:
            self.rpars.SEARCH_MAX_DGEN[target] = max_dgen
        if should_store_new_value and is_updating:
            self.rpars.searchConvInit['dgen'][target] = max_dgen

        if scaling is not None and scaling >= 1:
            self.rpars.SEARCH_MAX_DGEN_SCALING[target] = scaling
        elif scaling:
            message = 'dgen scaling value cannot be smaller than 1'
            raise ParameterRangeError(param, message=message)

    def _interpret_search_convergence_gaussian(self, param, gauss_width,
                                               scaling, is_updating):
        """Assign the GAUSSIAN_WIDTH/_SCALING parameters."""
        should_store_new_value = (
            not is_updating
            or gauss_width != self.rpars.searchConvInit['gaussian']
            )
        valid = gauss_width is not None and gauss_width > 0
        if should_store_new_value and not valid:
            message = 'gaussian width should be a positive number'
            raise ParameterRangeError(param, message=message)

        if should_store_new_value:
            self.rpars.GAUSSIAN_WIDTH = gauss_width
        if should_store_new_value and is_updating:
            self.rpars.searchConvInit['gaussian'] = gauss_width

        if scaling is not None and 0 < scaling <= 1:
            self.rpars.GAUSSIAN_WIDTH_SCALING = scaling
        elif scaling is not None:
            raise ParameterRangeError(param, scaling, (0, 1))

    def interpret_search_cull(self, assignment):
        """Assign parameter SEARCH_CULL."""
        param = 'SEARCH_CULL'
        rpars = self.rpars
        if len(assignment.values) not in (1, 2):
            rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param)
        try:
            cull = rpars.SEARCH_CULL.from_value(assignment.values)
        except TypeError as exc:   # Numeric value is not a number
            rpars.setHaltingLevel(1)
            raise ParameterFloatConversionError(param,
                                                message=str(exc)) from exc
        except ValueError as exc:  # Numeric or type are wrong
            rpars.setHaltingLevel(1)
            raise ParameterValueError(param, message=str(exc)) from exc

        if not cull.has_type:
            _default = rpars.get_default('SEARCH_CULL')
            cull.type_ = _default.type_
        rpars.SEARCH_CULL = cull

    def interpret_search_population(self, assignment):
        param = 'SEARCH_POPULATION'
        self.interpret_numerical_parameter(
            assignment,
            bounds=NumericBounds(type_=int, range_=(1, None))
            )
        if self.rpars.SEARCH_POPULATION < 16:
            _LOGGER.warning(f'{param} is very small. A minimum '
                            'value of 16 is recommended.')

    def interpret_search_start(self, assignment):
        param = 'SEARCH_START'
        # there should only be one values
        if assignment.other_values:
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param)

        # there are only a few options, so we can just check them all
        search_start = assignment.value
        if search_start.startswith('rand'):
            self.rpars.SEARCH_START = 'random'
        elif search_start.startswith('center'):
            self.rpars.SEARCH_START = 'centered'
        elif search_start.startswith('control'):
            self.rpars.SEARCH_START = 'control'
        elif search_start.startswith('crandom'):
            self.rpars.SEARCH_START = 'crandom'
        else:
            self.rpars.setHaltingLevel(1)
            raise ParameterUnknownFlagError(param, search_start)

    def interpret_site_def(self, assignment):                                   # TODO: Issue #143
        """Assign the SITE_DEF for one POSCAR element."""
        param = 'SITE_DEF'
        if assignment.values_str.count('top(') > 1:
            self.rpars.setHaltingLevel(3)
            raise ParameterValueError(
                param,
                message='only a single top() allowed per SITE_DEF line'
                )
        self._ensure_single_flag_assignment(assignment)
        site_element = self._get_valid_slab_element_from_flag(param,
                                                              assignment)
        sorted_atoms = []
        if 'top(' in assignment.values_str:
            sorted_atoms = sorted(
                (at for at in self.slab if at.el == site_element),
                key=lambda atom: atom.pos[2],
                reverse=True
                )

        site_def_dict = {}
        for flag_and_values in assignment.values_str.strip().split(','):
            site_label, site_specs = self._get_valid_sitedef_spec(
                param, site_element, flag_and_values
                )
            try:
                atnums = self._get_atom_numbers_for_site(site_specs,
                                                         sorted_atoms)
            except ValueError as exc:  # Invalid syntax
                self.rpars.setHaltingLevel(3)
                err_ = f'Invalid syntax in {assignment.values_str!r}: {exc!r}'
                raise ParameterParseError(param, message=err_) from None
            try:
                self._ensure_sitedef_makes_sense(site_element,
                                                 site_label,
                                                 atnums)
            except ValueError as exc:
                self.rpars.setHaltingLevel(3)
                raise ParameterValueError(assignment.parameter,
                                          message=str(exc)) from None
            site_def_dict[site_label] = atnums
        self.rpars.SITE_DEF[site_element] = site_def_dict

    def _get_valid_sitedef_spec(self, param, site_element, label_and_specs):
        """Return a site label and atom selection specifications, or raise."""
        try:
            site_label, *site_specs = (s.strip()
                                       for s in label_and_specs.split())
        except ValueError:
            self.rpars.setHaltingLevel(2)
            err_ = ('Expected "site_label atom_selection_patterns ". '
                    f'Found "{label_and_specs}"')
            raise ParameterNumberOfInputsError(param, message=err_) from None
        if not site_specs:
            self.rpars.setHaltingLevel(2)
            err_ = ('No atom selection pattern found for site '
                    f'label {site_label}, element {site_element}.')
            raise ParameterParseError(param, message=err_)
        return site_label, site_specs

    @staticmethod
    def _get_atom_numbers_for_site(site_specs, sorted_atoms):
        """Return atom numbers given a selection specification.

        Parameters
        ----------
        site_specs : Sequence
            Items are strings. Each corresponds to a specification for
            selecting some atom numbers. Acceptable specifications are:
            - single integer
            - range of integers (e.g., '1-5', '10:28')
            - 'top(N)'
        sorted_atoms : Sequence
            Atom objects, with element already correct for the specific
            site, sorted by decreasing fractional position. Used only
            if one of the `site_specs` is 'top(N)'.

        Returns
        -------
        atnums : list
            Atom numbers that are selected by the given `site_specs`.

        Raises
        ------
        ValueError
            If any of the `site_specs` has invalid syntax.
        """
        atnums = []
        for site_spec in site_specs:
            extra_atom_numbers = readIntRange(site_spec)
            if extra_atom_numbers:
                atnums.extend(extra_atom_numbers)
                continue
            if 'top(' in site_spec:
                n_top = int(site_spec.split('(')[1].split(')')[0])
                atnums.extend(at.num for at in sorted_atoms[:n_top])
                continue
            raise ValueError(site_spec)
        return atnums

    def _ensure_sitedef_makes_sense(self, poscar_el, site_label, atom_nrs):
        """Raise ValueError if the atom_nrs selected for a site are wrong."""
        invalid_atom_nums = []
        invalid_elements = []
        for num in atom_nrs:
            try:
                atom = self.slab.atlist.get(num)
            except KeyError:
                invalid_atom_nums.append(num)
                continue
            if atom.el != poscar_el:
                invalid_elements.append(atom)
        if not invalid_atom_nums and not invalid_elements:
            return
        error = f'Invalid selection for {site_label}:'
        if invalid_atom_nums:
            error += f'\nAtom number(s) {invalid_atom_nums} not found in slab'
        if invalid_elements:
            error += (f'\nAtom(s) {[str(at) for at in invalid_elements]} '
                      f'do not have POSCAR element {poscar_el}')
        raise ValueError(error)

    def interpret_superlattice(self, assignment):
        """Assign parameter SUPERLATTICE (Wood's or matrix notation)."""
        param = 'SUPERLATTICE'
        self._interpret_wood_or_matrix(param, assignment)
        self.rpars.superlattice_defined = True

    def interpret_symmetry_bulk(self, assignment):
        """Assign parameter SYMMETRY_BULK."""
        param = 'SYMMETRY_BULK'

        # We accept mirrors with syntax "m[a b]", rotations with
        # syntax "rN" and plane groups. Here we only check that
        # the syntax is OK. We defer the complaints about invalid
        # plane group, directions, or rotation orders, to when we
        # know more about the slab
        self.rpars.SYMMETRY_BULK = {  # 'group' added below
            'mirror': set(),
            'rotation': set()
            }
        unrecognized = assignment.values_str
        unrecognized = self._interpret_symmetry_bulk_glides(unrecognized)
        unrecognized = self._interpret_symmetry_bulk_screws(unrecognized)
        unrecognized = self._interpret_symmetry_bulk_group(unrecognized)

        if unrecognized:
            self.rpars.setHaltingLevel(2)
            message = f'Could not recognize {unrecognized!r}: '
            if 'm' in unrecognized:
                message += ('Syntax for mirrors is "m[n1 n2]"; '
                            'n1 and n2 must be 0, +-1, or +-2; ')
            if 'r' in unrecognized:
                message += ('Syntax for rotations is rN; N must be '
                            '2, 3, 4, or 6')
            raise ParameterValueError(param, message=message)

    def _interpret_symmetry_bulk_glides(self, unrecognized):
        """Look up `unrecognized` for glide-plane-direction specifications."""
        _mirror_re = re.compile(r'(\s+m\[\s*(-?[012])\s*,?\s*(-?[012])\])',
                                re.IGNORECASE)
        for token, *directions in _mirror_re.findall(unrecognized):
            # All matches of mirrors are acceptable
            unrecognized = unrecognized.replace(token, '')
            direction = tuple(int(v) for v in directions)
            if (direction[0] < 0
                    or not direction[0] and direction[1] < 0):
                direction = -direction[0], -direction[1]
            self.rpars.SYMMETRY_BULK['mirror'].add(direction)
        return unrecognized

    def _interpret_symmetry_bulk_group(self, unrecognized):
        """Look up `unrecognized` for a symmetry group specification."""
        param = 'SYMMETRY_BULK'
        _group_re = re.compile(                                                 # TODO: For now borrowed from guilib. Eventually will try to instantiate a PlaneGroup
            r'(\s*(\w+)\s*(?:\[\s*-?[012]\s*-?[012]\s*\])?)',
            re.IGNORECASE
            )
        for token, group in _group_re.findall(unrecognized):
            if group not in self.grouplist:
                continue
            unrecognized = unrecognized.replace(token, '')
            if 'group' in self.rpars.SYMMETRY_BULK:
                message = 'Only one symmetry group can be given'
                raise ParameterValueError(param, message=message)
            self.rpars.SYMMETRY_BULK['group'] = token.strip().lower()

        if 'group' not in self.rpars.SYMMETRY_BULK:
            message = 'Need to specify exactly one symmetry group'
            raise ParameterValueError(param, message=message)
        return unrecognized

    def _interpret_symmetry_bulk_screws(self, unrecognized):
        """Look up `unrecognized` for screw-axis-order specifications."""
        _rotation_re = re.compile(r'(\s+r([2346]))', re.IGNORECASE)
        for token, order in _rotation_re.findall(unrecognized):
            unrecognized = unrecognized.replace(token, '')
            self.rpars.SYMMETRY_BULK['rotation'].add(int(order))
        return unrecognized

    def interpret_symmetry_cell_transform(self, assignment):
        """Assign parameter SYMMETRY_CELL_TRANSFORM (Wood's or matrix)."""
        param = 'SYMMETRY_CELL_TRANSFORM'
        self._interpret_wood_or_matrix(param, assignment)

    def interpret_symmetry_eps(self, assignment):
        """Assign parameters SYMMETRY_EPS and SYMMETRY_EPS.z."""
        param = 'SYMMETRY_EPS'
        self._ensure_no_flags_assignment(assignment)
        if len(assignment.values) not in (1, 2):
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param)

        # warning specific to SYMMETRY_EPS
        warning_str = ('PARAMETERS file: SYMMETRY_EPS:\n'
                       'Given value {}is greater than one Ångström. This is a '
                       'very loose constraint and might lead to incorrect '
                       'symmetry detection. Be sure to check the output.')

        # interpret first value as SYMMETRY_EPS
        bounds = NumericBounds(type_=float, range_=(1e-100, None))
        eps_value = self.interpret_numerical_parameter(assignment,
                                                       bounds=bounds,
                                                       return_only=True)
        if eps_value > 1.0:
            # pylint: disable-next=logging-format-interpolation
            _LOGGER.warning(warning_str.format(''))
        # interpret possible second value as SYMMETRY_EPS.z
        if not assignment.other_values:
            self.rpars.SYMMETRY_EPS = SymmetryEps(eps_value)
            return

        z_assignment = Assignment(assignment.other_values, param)
        eps_z_value = self.interpret_numerical_parameter(z_assignment,
                                                         bounds=bounds,
                                                         return_only=True)
        self.rpars.SYMMETRY_EPS = SymmetryEps(eps_value, eps_z_value)
        if self.rpars.SYMMETRY_EPS.z > 1.0:
            # pylint: disable-next=logging-format-interpolation
            _LOGGER.warning(warning_str.format('for z '))

    def interpret_symmetry_fix(self, assignment):                               # TODO: use symmetry groups from elsewhere once symmetry and guilib are merged
        param = 'SYMMETRY_FIX'
        group = assignment.values_str.lower()
        if group.startswith('t'):  # determine symmetry automatically
            self.rpars.reset_default(param)
            return
        if group.startswith('f'):
            self.rpars.SYMMETRY_FIX = 'p1'
            return
        if group in self.grouplist and group in ('cm', 'pmg'):
            message = f'For group {group} direction needs to be specified'
            self.rpars.setHaltingLevel(1)
            raise ParameterParseError(param, message)
        if group in self.grouplist:
            self.rpars.SYMMETRY_FIX = group
            return
        if group.startswith(('pm', 'pg', 'cm', 'rcm', 'pmg')):
            # regex to read
            rgx = re.compile(
                r'\s*(?P<group>(pm|pg|cm|rcm|pmg))\s*'
                + r'\[\s*(?P<i1>[-012]+)\s+(?P<i2>[-012]+)\s*\]')
            m = rgx.match(assignment.values_str.lower())
            if not m:
                self.rpars.setHaltingLevel(2)
                raise ParameterParseError(param)
            i1 = i2 = -2
            group = m.group('group')
            try:
                i1 = int(m.group('i1'))
                i2 = int(m.group('i2'))
            except ValueError as err:
                self.rpars.setHaltingLevel(2)
                raise ParameterParseError(param) from err
            if (group in ['pm', 'pg', 'cm', 'rcm', 'pmg']
                    and i1 in range(-1, 3) and i2 in range(-1, 3)):
                self.rpars.SYMMETRY_FIX = m.group(0)
            else:
                self.rpars.setHaltingLevel(2)
                raise ParameterParseError(param)
        else:
            self.rpars.setHaltingLevel(2)
            raise ParameterParseError(param)

    def interpret_tensor_output(self, assignment):
        """Assign parameter TENSOR_OUTPUT."""
        param = 'TENSOR_OUTPUT'
        tensor_output = []
        # Acceptable format for each token is (repeats*)value,
        # with the bracketed part as optional and 'value' either
        # zero or one (or a bool that can be translated to those)
        rgx = re.compile(r'\s*((?P<repeats>\d+)\s*[*])?\s*(?P<value>[01])\s*$')
        tokens = recombineListElements(assignment.values, '*')
        for token in tokens:
            token_01 = re.sub(r'[Tt](rue|RUE)?', '1', token)
            token_01 = re.sub(r'[Ff](alse|ALSE)?', '0', token_01)

            _match = rgx.match(token_01)
            if not _match:
                self.rpars.setHaltingLevel(1)
                raise ParameterParseError(param,
                                          f'Invalid format in: {token!r}')
            repeats = int(_match['repeats'] or '1')
            value = int(_match['value'])
            tensor_output.extend((value,)*repeats)
        self.rpars.TENSOR_OUTPUT = tensor_output

    def interpret_theo_energies(self, assignment):
        """Assign parameter THEO_ENERGIES. Correct start if inappropriate."""
        param = 'THEO_ENERGIES'
        energies = assignment.values

        # (1) Single value input: treat as default if it is a single
        # underscore, otherwise as a request for a single energy
        if not assignment.other_values and assignment.value == '_':
            self.rpars.reset_default(param)
            return
        if not assignment.other_values:
            energies = (assignment.value, assignment.value, '1')

        # (2) Three values. Any can be an "_" meaning "default".
        # Internally, we store those as an rparams.NO_VALUE. The
        # others should be positive floats.
        theo_energies = self._parse_energy_range(assignment, energies)
        if len(theo_energies) != 3:
            raise ParameterNumberOfInputsError(param)

        try:
            self.rpars.THEO_ENERGIES = TheoEnergies(*theo_energies)
        except ValueError as exc:  # No TypeError, as all are floats
            self.rpars.setHaltingLevel(1)
            raise ParameterValueError(param, message=str(exc)) from None

        # Notify in case start was adjusted to fit step
        if not self.rpars.THEO_ENERGIES.is_equivalent(theo_energies):
            _LOGGER.info(
                f'{param} parameter: (Eto - Efrom) % Estep != 0, Efrom '
                f'was corrected to {self.rpars.THEO_ENERGIES.start}'
                )

    def interpret_tl_version(self, assignment):
        """Assign parameter TL_VERSION."""
        param = 'TL_VERSION'
        self._ensure_simple_assignment(assignment)
        version_str = assignment.value.lower()
        version_str = version_str.replace('v', '')
        if version_str in OLD_TL_VERSION_NAMES.keys():
            version_str = OLD_TL_VERSION_NAMES[version_str]
        try:
            self.rpars.TL_VERSION = Version(version_str)
        except ValueError:
            raise ParameterConversionError(param, version_str)

    def interpret_v0_real(self, assignment):
        """Assign parameter V0_REAL."""
        param = 'V0_REAL'
        v0r_type = assignment.value.lower()
        if v0r_type == 'rundgren':
            rundgren_constants = assignment.other_values
            if len(rundgren_constants) != 4:
                message = ('Rundgren-type function expects four '
                           'constants separated by whitespace')
                raise ParameterNumberOfInputsError(param, message=message)
            try:
                self.rpars.V0_REAL = [float(c) for c in rundgren_constants]
            except ValueError as exc:
                message = (f'Could not parse constants {rundgren_constants} '
                           'for Rundgren-type function')
                self.rpars.setHaltingLevel(1)
                raise ParameterFloatConversionError(param,
                                                    message=message) from exc
            return

        # Pass a specific function to FORTRAN, but replace
        # 'EE' with 'EEV+workfn' to get the energies right
        v0_real = re.sub('(?i)EE', 'EEV+workfn', assignment.values_str)
        self.rpars.V0_REAL = v0_real.rstrip()

    def interpret_vibr_amp_scale(self, assignment):
        """Store raw values: Requires sites for interpreting."""
        self.rpars.VIBR_AMP_SCALE.extend(assignment.values_str.split(','))

    def _interpret_wood_or_matrix(self, param, assignment):
        """Store a Wood- or matrix-notation parameter value."""
        if 'M' not in assignment.flags:
            self._ensure_no_flags_assignment(assignment)
            wood = self._read_woods_notation(param, assignment)
            setattr(self.rpars, param, wood)
        else:
            matrix = self._read_matrix_notation(param, assignment.values)
            setattr(self.rpars, param, matrix)

    # ------------------------ Helper methods -------------------------
    def _ensure_chemical_elements(self, param, elements):
        """Raise unless all entries are valid chemical elements."""
        if not elements:
            raise ParameterHasNoValueError(param)
        invalid = []
        for element in elements:
            try:
                _ = periodic_table.get_atomic_number(element)
            except ValueError:
                invalid.append(element)

        if invalid:
            message = f'Element(s) {invalid} not found in periodic table'
            self.rpars.setHaltingLevel(2)
            raise ParameterValueError(param, message=message)

    def _ensure_has_non_empty_slab(self, param):
        """Complain if self.slab is inappropriate for SITE_DEF."""
        if self.slab is None:
            self.rpars.setHaltingLevel(3)
            raise ParameterNeedsSlabError(param)
        if not self.slab.atlist:
            self.rpars.setHaltingLevel(3)
            raise ParameterError(param, 'Slab contains no atoms')

    def _ensure_valid_slab_element(self, param, element):
        """Raise unless element is one of the elements of self.slab."""
        self._ensure_has_non_empty_slab(param)
        known_elements = self.slab.elements
        if not known_elements:
            self.rpars.setHaltingLevel(3)
            raise ParameterError(param, 'Slab has no elements')
        if element not in known_elements:
            msg = (f'{element!r} is not a valid POSCAR element. '
                   'Known elements: ' + ', '.join(known_elements))
            self.rpars.setHaltingLevel(3)
            raise ParameterUnknownFlagError(param, message=msg)

    def _get_flag_value_from_pair(self, param, flag_value_pair):
        """Return lowercase flag and value from a pair, or complain."""
        try:
            flag, value = (s.strip()
                           for s in flag_value_pair.lower().strip().split())
        except ValueError:
            err_ = f'Expected "flag value" pairs, found "{flag_value_pair}"'
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(param, message=err_) from None
        return flag, value

    def _get_valid_slab_element_from_flag(self, param, assignment):
        """Return a valid POSCAR element from assignment; raise otherwise."""
        element = assignment.flag.capitalize()
        self._ensure_valid_slab_element(param, element)
        return element

    def _parse_energy_range(self, assignment, energies):
        """Return a list of floats for energies.

        Parameters
        ----------
        assignment : Assignment
            The assignment for the parameter. Used only for error
            reporting.
        energies : Sequence
            The energies that should be parsed. Elements are strings.
            They may also contain single underscore characters, which
            will be replaced with an Rparams.NO_VALUE.

        Returns
        -------
        float_energies : list
            The energies as a list of floats

        Raises
        ------
        ParameterFloatConversionError
            If conversion to float fails for some of the inputs
        ParameterParseError
            If other parsing errors occur (typically a malformed
            input that cannot be converted to a simple tuple of
            of numbers).
        """
        try:
            return EnergyRange.parse_string_sequence(energies)
        except (TypeError, ValueError) as exc:
            self.rpars.setHaltingLevel(1)
            new_exc = (ParameterParseError if isinstance(exc, ValueError)
                       else ParameterFloatConversionError)
            raise new_exc(assignment.parameter,
                          message=assignment.values_str) from exc

    def _parse_incidence_angles(self, param, assignment):                       # TODO: Issue #147
        """Return a dictionary with incidence angles from an assignment."""
        bounds = {'THETA': NumericBounds(type_=float, range_=(-90, 90),
                                         out_of_range_event='fail'),
                  'PHI': NumericBounds(type_=float, range_=(0, 360),
                                       out_of_range_event='modulo')}
        if ',' not in assignment.values_str and len(assignment.values) != 2:
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(
                parameter=param,
                found_and_expected=(len(assignment.values), 2)
                )
        right_side = assignment.values_str.upper()
        if ',' not in assignment.values_str:
            # pylint: disable=consider-using-f-string
            # disable: Better than repeating 'assignment.values' twice
            right_side = 'THETA {}, PHI {}'.format(*assignment.values)
        self._check_n_incidence_angles_ok(param, right_side)

        angles_specs = (spec.strip().split()
                        for spec in right_side.strip().split(','))
        angles = {}
        for name, *values in angles_specs:
            if name not in ('THETA', 'PHI'):
                self.rpars.setHaltingLevel(1)
                raise ParameterUnknownFlagError(param, name)
            if len(values) != 1:
                self.rpars.setHaltingLevel(1)
                raise ParameterNumberOfInputsError(
                    parameter=param,
                    message=f'Found {len(values)} values for angle {name}'
                    )
            angles[name] = self.interpret_numerical_parameter(
                Assignment(values[0], param),
                param=f'{param} {name}',
                bounds=bounds[name],
                return_only=True
                )

        # Check for negative theta and adjust phi
        if angles['THETA'] < 0:
            angles['THETA'] = abs(angles['THETA'])
            angles['PHI'] = (angles['PHI'] + 180) % 360
        return angles

    def _check_n_incidence_angles_ok(self, param, right_side):
        """Complain if right_side contains the wrong number of angles."""
        nr_angle_specs = len(right_side.split(','))
        if nr_angle_specs != 2:
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(
                parameter=param, found_and_expected=(nr_angle_specs, 2)
                )
        if (right_side.count('THETA'), right_side.count('PHI')) != (1, 1):
            self.rpars.setHaltingLevel(1)
            raise ParameterNumberOfInputsError(
                parameter=param,
                message='Expected exactly one entry for THETA and one for PHI'
                )

    def _read_woods_notation(self, param, assignment):
        """Return a Woods notation from an Assignment, if slab exists."""
        if self.slab is None:
            message = (f'{param} parameter appears to be in Wood '
                       'notation but no slab was passed. Cannot '
                       'calculate bulk unit cell!')
            self.rpars.setHaltingLevel(2)
            raise ParameterNeedsSlabError(param, message)
        return readWoodsNotation(assignment.values_str, self.slab.ucell)

    def _read_matrix_notation(self, param, values):
        """Try interpreting values as a 2x2 matrix."""
        matrix = splitSublists(values, ',')
        if len(matrix) != 2:
            message = 'Number of lines in matrix is not equal to 2'
            self.rpars.setHaltingLevel(2)
            raise ParameterParseError(param, message)
        if any(len(row) != 2 for row in matrix):
            message = 'Number of columns in matrix is not equal to 2'
            self.rpars.setHaltingLevel(2)
            raise ParameterParseError(param, message)
        try:
            return np.array(matrix).astype(float)
        except ValueError:
            self.rpars.setHaltingLevel(1)
            raise ParameterFloatConversionError(param, matrix) from None



"""Python extensions written in FORTRAN."""

import warnings

try:
    from .rfactor import r_factor_new as rf
    from .interpolation import interpolation as intpol
except ImportError:
    pass # suppress warnings - they are very annoying
    # warnings.warn("F2PY compiled module rfactor and/or interpolation failed to import.")
    # warnings.warn("Check that these modules were compiled with the correct version of numpy/F2PY.")



"""Module sym_entity of viperleed.calc.classes.

Part of the functionality defined here used to be in the slab module.
Defines the SymPlane class, which represent a mirror/glide operations
with a 2D location.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-16'
__license__ = 'GPLv3+'

import numpy as np

from viperleed.calc.lib.base import add_edges_and_corners, dist_from_line


class SymPlane:
    """Candidate plane for a symmetry operation. 'ty' pre-defines a type
    (mirror or glide), 'index2' allows the (1,2) and (2,1) directions if True,
    and collapse moves pos into the (0,0) unit cell if True."""

    def __init__(self, pos, dr, abt, ty="none", index2=False, collapse=True):
        if collapse:  # collapse to (0,0) cell
            self.pos = np.dot(abt.T, (np.dot(np.linalg.inv(abt.T), pos) % 1.0))
        else:
            self.pos = pos
        self.dir = dr/np.linalg.norm(dr)
        # normalized vector perpendicular to pos = in-plane
        self.type = ty
        self.par = []
        optionlist = [(1, 0), (0, 1), (1, 1), (1, -1)]
        if index2:
            optionlist.extend([(2, 1), (1, 2)])
        for (i, j) in optionlist:
            if abs((abs(np.dot(self.dir, (i*abt[0]+j*abt[1])))
                    / (np.linalg.norm(self.dir)
                    * np.linalg.norm(i*abt[0]+j*abt[1])))-1.0) < 0.001:
                self.par = np.array([i, j])

    def __str__(self):
        """Return a string representation of this SymPlane."""
        return f'SymPlane(pos={self.pos}, par={self.par})'

    @property
    def is_glide(self):
        """Return whether this is a mirror plane."""
        return self.type == 'glide'

    @property
    def is_mirror(self):
        """Return whether this is a mirror plane."""
        return self.type == 'mirror'

    @property
    def normal(self):
        """Return a unit vector normal to this plane (even without perp)."""
        return np.array((self.dir[1], -self.dir[0]))

    def distanceFromOrigin(self, abt):
        pointlist = [(0, 0), (1, 0), (0, 1), (1, 1)]
        return min([dist_from_line(self.pos, self.pos+self.dir,
                                   p[0]*abt[0]+p[1]*abt[1])
                    for p in pointlist])

    def isEquivalent(self, pl2, abt, eps=0.001):
        """Checks whether two symmetry planes have the same position and
        direction (including duplicates in next unit cell)"""
        if not np.array_equal(self.par, pl2.par):
            return False

        # If we're close to an edge or corner, also check translations
        fpos = np.dot(np.linalg.inv(np.transpose(abt)), self.pos) % 1.0
        releps = eps / np.linalg.norm(abt, axis=1)
        complist, _ = add_edges_and_corners([self.pos], (fpos,), releps, abt)

        for p in complist:
            if dist_from_line(pl2.pos, pl2.pos+pl2.dir, p) < eps:
                return True
        return False

    def point_operation(self, n_dim=2):
        """Return a matrix for mirroring across this plane."""
        # See https://en.wikipedia.org/wiki/Householder_transformation
        normal = self.normal
        mirror = np.identity(n_dim)
        mirror[:2, :2] -= 2*np.outer(normal, normal)
        return mirror



"""Functions for reading and writing beams files.

This includes: BEAMLIST, IVBEAMS, EXPBEAMS and AUXEXPBEAMS.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import copy
from io import StringIO
import logging
import os
import re

import fortranformat as ff
import numpy as np

from viperleed.calc.classes.beam import Beam
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import parseMathSqrt
from viperleed.guilib import project_to_first_domain


logger = logging.getLogger(__name__)


def averageBeams(beams, weights=None):
    """Takes a list of percentages and a list of lists of Beam objects.
    Returns a new list of Beam objects with weighted averaged intensities."""
    if beams is None:
        raise ValueError("averageBeams: No beams passed.")
    if weights is None or len(weights) == 0:
        weights = [1/len(beams)] * len(beams)
    avbeams = copy.deepcopy(beams[0])
    for (i, b) in enumerate(avbeams):
        if not all([beams[j][i].isEqual(b) for j in range(1, len(beams))]):
            _err = "averageBeams: Beam lists are mismatched."
            logger.error(_err)
            raise ValueError(_err)
            return []
        if not all([set(beams[j][i].intens.keys()) == set(b.intens.keys())
                    for j in range(1, len(beams))]):
            _err = "averageBeams: Beams have different energy ranges."
            logger.error(_err)
            raise ValueError(_err)
            return []
        for en in b.intens:
            b.intens[en] *= weights[0]
            for j in range(1, len(beams)):
                b.intens[en] += beams[j][i].intens[en] * weights[j]
    return avbeams


def readBEAMLIST(filename="BEAMLIST"):
    """Reads the BEAMLIST file and returns the contents as a list of the
    lines as strings."""
    beamlist = []
    try:
        with open(filename, "r") as rf:
            beamlist = rf.readlines()
        logger.debug("BEAMLIST file was read successfully")
    except Exception:
        logger.error("Error opening BEAMLIST file.")
        raise
    return beamlist


def readIVBEAMS(filename='IVBEAMS'):
    """Reads an IVBEAMS file and returns a list of beams (using Beam class)"""
    # open input file
    try:
        with open(filename, 'r') as rf:
            ivbeamlines = rf.readlines()
    except Exception:
        raise
    linenum = 1		# iterates the current line being read
    hklist = []
    # TODO: skip commented out lines (same as PARAMETERS, etc.)
    for line in ivbeamlines:
        # ignore brackets and vbars, except as spacers
        line = line.replace("(", " ")
        line = line.replace(")", " ")
        line = line.replace("|", " ")
        llist = line.split()
        if len(llist) == 1 and linenum != 1:
            logger.warning('A line with only one element was found in '
                           'IVBEAMS and will be skipped: '+line)
        elif len(llist) >= 2:
            f = [None, None]
            for i in range(0, 2):
                try:
                    f[i] = float(llist[i])
                except ValueError:
                    if '/' in llist[i]:
                        try:
                            f[i] = (float(llist[i].split('/')[0])
                                    / float(llist[i].split('/')[1]))
                        except ValueError:
                            if linenum != 1:
                                logger.error('Error reading IVBEAMS line: '
                                             + line)
                                raise
                    else:
                        if linenum != 1:
                            logger.error('Error reading IVBEAMS line: '+line)
                            raise
            if linenum != 1:
                if not (f[0], f[1]) in hklist and None not in f:
                    hklist.append((f[0], f[1]))
            else:
                # check whether there is data in first line by mistake
                if None not in f:  # data was read
                    logger.warning(
                        'It looks like the first line in the '
                        'IVBEAMS file may contain data. Note that the first '
                        'line should be a header line, so the data will not '
                        'be read.')
        linenum += 1
    beams = []
    for hk in hklist:
        beams.append(Beam(hk))
    logger.debug("IVBEAMS file was read successfully")
    return beams


def sortIVBEAMS(sl, rp):
    """Sorts the beams in IVBEAMS such that they appear in the same order as
    in the BEAMLIST. Returns the sorted list."""
    # read BEAMLIST
    if rp.beamlist == []:
        logger.warning("sortIVBEAMS routine: no beamlist passed, "
                       "attempting to read BEAMLIST directly.")
        try:
            with open('BEAMLIST', 'r') as rf:
                rp.beamlist = rf.readlines()
        except FileNotFoundError:
            logger.error("BEAMLIST not found.")
            raise
    err = 1e-3          # since beams are saved as floats, give error tolerance
    symeq = leedbase.getSymEqBeams(sl, rp)
    # first, get beamlist as floats
    blfs = []
    for line in rp.beamlist:
        llist = line.split()
        if len(llist) > 1:
            fl = []
            for i in range(0, 2):
                try:
                    fl.append(float(llist[i]))
                except (ValueError, IndexError):
                    pass
            if len(fl) == 2:
                blfs.append(fl)
    # now figure out if there are beams in IVBEAMS without a partner
    not_found = []
    found_equivalent = []
    for (ind, ib) in enumerate(rp.ivbeams):
        found = any([ib.isEqual_hk(lb, eps=err) for lb in blfs])
        if found:
            continue
        eqbl = []   # hk of equivalent beams
        for (hk, i) in symeq:
            if ib.isEqual_hk(hk, eps=err):
                eqbl.extend([hk2 for (hk2, j) in symeq if i == j])
        for eqb in eqbl:
            for lb in blfs:
                if abs(eqb[0]-lb[0]) < err and abs(eqb[1]-lb[1]) < err:
                    # rename the beam
                    if any([ib2.isEqual_hk(eqb, eps=err)
                            for ib2 in rp.ivbeams]):
                        found_equivalent.append(ib.label)
                    else:
                        b = Beam(eqb)
                        logger.debug(
                            "Beam " + ib.label + " is not in "
                            "BEAMLIST, renaming to equivalent beam "
                            + b.label + ".")
                        rp.ivbeams[ind] = b
                    found = True
                    break
            if found:
                break
        if not found:
            not_found.append(ib.label)
    if found_equivalent:
        logger.debug(
            "The following beams from IVBEAMS will be dropped because they "
            "are not in BEAMLIST, but equivalent beams already are: "
            + ", ".join(found_equivalent))
    if not_found:
        logger.warning(
            "The following beams from IVBEAMS are not in BEAMLIST and will be "
            "dropped: " + ", ".join(not_found))
    # now sort
    ivsorted = []
    for lb in blfs:
        ivsorted.extend([ib for ib in rp.ivbeams if ib.isEqual_hk(lb,
                                                                  eps=err)])
    return ivsorted


def readOUTBEAMS(filename="EXPBEAMS.csv", sep=",", enrange=None):
    #TODO: this need some optimizing; use csv and dict reader class
    """Reads beams from an EXPBEAMS.csv or THEOBEAMS.csv file.
    
    The beams are returned as a list of Beam objects.
    
    Parameters
    ----------
    filename : str, StringIO, optional
        If string, interpret as the path to the file to read. If
        a StringIO object, assume it contains already the contents
        of the file to be interpreted. Default is "EXPBEAMS.csv".
    sep: str, optional
        Separator character for splitting beams. Default is ",".
    enrange : Sequence, or None, optional
        Energy range to be used to filter beams. Beams that contain no
        data within the range will be filtered out before returning.
        If given and not None, it should be a Sequence with two elements
        (min and max).
    
    Returns
    -------
    beams : list of Beam
        Beam objects with info read from the input given.
    """
    beams = []
    # Deals with input in the form of a StingIO (can be used to feed file input as a string)
    if isinstance(filename, StringIO):
        lines = (li[:-1] for li in filename)
        filename = "StringIO"
    else:
        try:
            with open(filename, 'r') as rf:
                lines = [li[:-1] for li in rf.readlines()]
        except FileNotFoundError:
            if filename.endswith(".csv") and os.path.isfile(filename[:-4]):
                with open(filename[:-4], 'r') as rf:
                    lines = [li[:-1] for li in rf.readlines()]
            else:
                logger.error("Error reading "+filename)
                raise

    firstline = True
    rgx = re.compile(r'[\*\(\s]*(?P<h>[-0-9/]+)\s*\|\s*(?P<k>[-0-9/]+)')
    for line in lines:
        if firstline and sep not in line:   # try some other separators
            for sep2 in [s for s in [",", ";"] if s != sep]:
                if sep2 in line and len(line.split(sep2)) > 2:
                    logger.info("Found separator '"+sep2+"' in "+filename
                                + ", expected '"+sep+"'. Attempting to read "
                                "with '"+sep2+"'.")
                    sep = sep2
                    break
        llist = line.split(sep)
        if llist[-1].strip() == "":
            llist = llist[:-1]
        if firstline:
            firstline = False
            for label in llist[1:]:
                m = rgx.match(label)
                if m is None:
                    logger.error("readOUTBEAMS: Could not parse h/k in "
                                 "label: "+label)
                    return []
                sh = m.group("h")   # string h
                sk = m.group("k")   # string k
                try:
                    if "/" in sh: # TODO if fractionals found - pass fractions already to beam
                        h = int(sh.split("/")[0]) / int(sh.split("/")[1])
                    else:
                        h = float(sh)
                    if "/" in sk:
                        k = int(sk.split("/")[0]) / int(sk.split("/")[1])
                    else:
                        k = float(sk)
                except (ValueError, IndexError, ZeroDivisionError):
                    logger.error("readOUTBEAMS: Could not parse h/k in "
                                 "label: "+label)
                    return []
                beams.append(Beam((h, k)))
        elif len(line) > 1:
            try:
                en = float(llist[0])
            except ValueError:
                logger.error("readOUTBEAMS: Could not parse "+llist[0]+"as "
                             "an energy")
                return []
            for i in range(0, len(llist)):
                try:
                    f = float(llist[i+1])
                    if not np.isnan(f):
                        beams[i].intens[en] = f
                except (ValueError, IndexError):
                    pass
                
    # cleanup of beams
    threshold_intes = 1e-8  # threshold value of TensErLEED (rfacsb.f in subroutine grid)
    lowest_real_en = 0
    for beam in beams:

        unfiltered_intens = copy.deepcopy(list(beam.intens.items())) 
        # find first intensity > threshold
        for en, intensity in unfiltered_intens:
            if intensity > threshold_intes:
                lowest_real_en = en
                break

        # remove parts of beam that are set to NaN or below threshhold
        for en, intensity in unfiltered_intens:
            if (np.isnan(intensity) or (en < lowest_real_en)):
                beam.intens.pop(en)

    if enrange is not None and len(enrange) == 2:
        remlist = []
        for b in beams:
            if (len(b.intens) == 0 or
                    (enrange[0] > 0 and max(b.intens) < enrange[0]) or
                    (enrange[1] > 0 and min(b.intens) > enrange[1])):
                # beam has no data in given interval; remove
                remlist.append(b)
        if remlist:
            logger.warning("The following beams contain no data in the given "
                           "energy range and will be ignored: "
                           + ", ".join([b.label for b in remlist]))
        for b in remlist:
            beams.remove(b)

    # Add an offset such that intensities are always positive, and warn.
    for beam in beams:
        min_intensity = min((0, *beam.intens.values()))
        if min_intensity < 0:
            logger.warning(f"Negative intensity encountered in beam {beam.label} while reading {filename}."
                           f" An offset was added so that the minimum intensity of this beam is 0.")
            for en in beam.intens.keys():
                beam.intens[en] += abs(min_intensity)

    if enrange is None:
        return beams
    totalrange = 0.
    minmax = enrange[:]
    if len(minmax) < 2:
        minmax = [0, 1e5]
    if minmax[0] < 0:
        minmax[0] = 0
    if minmax[1] <= 0:
        minmax[1] = 1e5
    for b in beams:
        totalrange += (min(max(b.intens), minmax[1])
                       - max(min(b.intens), minmax[0]))
    logger.info("Loaded "+filename+" file containing {} beams (total energy "
                "range: {:.2g} eV).".format(len(beams), totalrange))

    return beams


def checkEXPBEAMS(sl, rp, domains=False):
    remlist = []
    symeq = leedbase.getSymEqBeams(sl, rp)
    for (bi, b) in enumerate(rp.expbeams):
        if b in remlist:
            continue
        eqbl = []   # hk of equivalent beams
        for (hk, i) in symeq:
            if b.isEqual_hk(hk):
                eqbl.extend([hk2 for (hk2, j) in symeq if i == j])
                break
        for b2 in rp.expbeams[bi+1:]:
            for hk in eqbl:
                if b2.isEqual_hk(hk):
                    remlist.append(b2)
                    w = ("Experimental beam "+b2.label+" is "
                         "equivalent to experimental beam "+b.label+". ")
                    if rp.HALTING > 2:
                        w += "Beam "+b2.label+" will be discarded."
                    else:
                        w += ("Remove one of them, or average them as "
                              "appropriate.")
                    logger.warning(w)
                    rp.setHaltingLevel(2)
                    break
    for b in remlist:
        rp.expbeams.remove(b)


def readAUXEXPBEAMS(filename="AUXEXPBEAMS", interactive=False):
    """Reads beams from an AUXEXPBEAMS file, which already has the formatting
    required by TensErLEED. Returns a list of Beam objects. Only works if the
    comment lines are of format *( h k )*, including the stars, with h and k
    separated by whitespace. If 'interactive' is True, then the user will be
    asked to clarify labels that cannot be read."""
    expbeams = []
    try:
        with open(filename, 'r') as rf:
            lines = [li[:-1] for li in rf.readlines()]
    except FileNotFoundError:
        logger.error("Error reading AUXEXPBEAMS.")
        raise
    read = False
    rf62x12 = ff.FortranRecordReader('12F6.2')
    rgx = re.compile(r'[\*\(\s]*(?P<h>[-0-9/]+)\s+(?P<k>[-0-9/]+)')
    for line in lines:
        if "*" in line:
            read = True
            topline = True  # next line contains number of beams and scaling
            failedToRead = False
            m = rgx.match(line)
            if m is not None:
                sh = m.group("h")   # string h
                sk = m.group("k")   # string k
                try:
                    h = parseMathSqrt(sh)
                    k = parseMathSqrt(sk)
                except Exception:
                    failedToRead = True
                else:
                    newbeam = Beam((h, k))
                    expbeams.append(newbeam)
            else:
                failedToRead = True
            if failedToRead:
                if not interactive:
                    logger.error("readAUXEXPBEAMS: Could not parse h/k in "
                                 "line:\n"+line)
                    return []
                else:
                    print("readAUXEXPBEAMS: Could not parse h/k in "
                          "line:\n"+line)
                    while True:
                        try:
                            hks = input("Please manually input the beam "
                                        "indices (space-separated): ")
                        except Exception:
                            return []
                        if hks.lower() in ["quit", "exit"]:
                            return []
                        if hks and len(hks.split()) > 1:
                            try:
                                h = parseMathSqrt(hks.split()[0])
                                k = parseMathSqrt(hks.split()[1])
                            except Exception:
                                print("Could not parse h/k")
                            else:
                                newbeam = Beam((h, k))
                                expbeams.append(newbeam)
                                break
        elif read and topline:
            topline = False
            llist = line.split()
            try:
                scaling = float(llist[1])
            except Exception:
                logger.error("readAUXEXPBEAMS: Could not parse number of "
                             "beams or scaling factor in line:\n"+line)
                return []
        elif read:
            newvals = rf62x12.read(line)
            i = 0
            while i+1 < len(newvals):
                if newvals[i] is None:
                    break
                newbeam.intens[newvals[i]] = newvals[i+1] / scaling
                i += 2
    return expbeams


def writeIVBEAMS(sl, rp, filename="IVBEAMS", domains=False):
    """Writes an IVBEAMS file based on rp.exbeams. Returns
    those beams in IVBEAMS form."""
    if not domains:
        d = [leedbase.getLEEDdict(sl, rp)]
    else:
        d = [leedbase.getLEEDdict(dp.sl, dp.rp) for dp in rp.domainParams]
    if any([v is None for v in d]):
        logger.error("Failed to write IVBEAMS")
        return []
    output = ("  h         k          Beams to calculate, automatically "
              "generated from EXPBEAMS\n")
    if rp.AVERAGE_BEAMS is not False:
        makebeams = project_to_first_domain([b.hkfrac for b in rp.expbeams],
                                            *d)
        ivbeams = [Beam(hk) for hk in makebeams]
    else:
        ivbeams = [Beam(b.hkfrac) for b in rp.expbeams]
    for b in ivbeams:
        output += "{: 10.6f} {: 10.6f}\n".format(b.hk[0], b.hk[1])
    try:
        with open(filename, "w") as wf:
            wf.write(output)
        logger.debug("Wrote IVBEAMS file successfully.")
    except Exception:
        logger.error("Exception while writing IVBEAMS file: ",
                     exc_info=True)
        raise
    return ivbeams


def writeOUTBEAMS(beams, filename="THEOBEAMS.csv", sep=", ",
                  which="intensity"):
    """Takes a list of Beam objects and writes them to a comma-separated
    file. Parameter 'which' defines what to write; set to 'amp_real' or
    'amp_imag' to write real and imaginary parts of complex amplitudes."""
    nan = "NaN"  # what to put when no value
    output = "E".rjust(7)+sep
    if which == "intensity":
        minwidth = 11
    else:
        minwidth = 17
    w = max(minwidth, len(beams[0].label))
    energies = []
    for b in beams:
        output += b.label.rjust(w)+sep
        if which == "intensity":
            energies.extend([k for k in b.intens if k not in energies])
        else:
            energies.extend([k for k in b.complex_amplitude
                             if k not in energies])
    output = output[:-len(sep)]
    output += "\n"
    energies.sort()
    for en in energies:
        output += '{:7.2f}'.format(en) + sep
        for b in beams:
            if which == "intensity":
                write_dict = b.intens
            else:
                write_dict = b.complex_amplitude
            if en in write_dict:
                if which == "intensity":
                    output += '{:0.5E}'.format(b.intens[en]).rjust(w) + sep
                else:
                    if which == "amp_real":
                        v = b.complex_amplitude[en].real
                    else:
                        v = b.complex_amplitude[en].imag
                    output += '{:0.10E}'.format(v).rjust(w) + sep
            else:
                output += nan.rjust(w) + sep
        output = output[:-len(sep)]
        output += "\n"
    try:
        with open(filename, 'w') as wf:
            wf.write(output)
    except OSError:
        logger.error("Failed to write "+filename)
        raise
    logger.debug("Wrote to "+filename+" successfully")
    return


def writeAUXBEAMS(ivbeams=None, beamlist=None, beamsfile='IVBEAMS',
                  readfile='BEAMLIST', writefile='AUXBEAMS', write=True):
    """"Reads from a BEAMLIST file (full list of beams for calculation),
    finds the beams listed in IVBEAMS (if not passed as a list 'beams', will
    attempt to call readIVBEAMS directly) and copies the corresponding lines
    to AUXBEAMS. Returns a list of the corresponding beam numbers in BEAMLIST.
    """
    if not readfile == "BEAMLIST":
        blstr = "Beam list (filename "+readfile+")"
    else:
        blstr = "BEAMLIST"

    if ivbeams is None:         # if 'ivbeams' is empty, try to fill it
        ivbeams = readIVBEAMS(beamsfile)

    output = '   1               IFORM\n'  # always use formatted input+output

    # read BEAMLIST
    if beamlist is None:
        logger.warning("writeAUXBEAMS routine: no beamlist passed, "
                       "attempting to read BEAMLIST directly.")
        try:
            beamlist = readBEAMLIST(readfile)
        except Exception:
            logger.error("Error getting beamlist not found.")
            raise
    # since beams are saved as floats, give some error tolerance
    err = 1e-4
    # read BEAMLIST - very little error handling here, I'm assuming
    #  the beamlists are safe; could be added later
    foundbeams = []
    numlist = []
    blocks = 0
    totalbeams = 0
    for line in beamlist:
        llist = line.split()
        if len(llist) > 1:
            totalbeams += 1
            for b in ivbeams:
                if b.isEqual_hk((float(llist[0]), float(llist[1])), eps=err):
                    output += line
                    foundbeams.append(b)
                    numlist.append(int(line.split('.')[-1]))
        else:
            blocks += 1
    not_found = [b.label for b in ivbeams if b not in foundbeams]
    if not_found:
        logger.warning('IVBEAMS contains beams that are not in ' + blstr + ": "
                       + ", ".join(not_found))
    if write:
        if not writefile == 'AUXBEAMS':
            wfstr = 'AUXBEAMS file (filename '+writefile+')'
        else:
            wfstr = 'AUXBEAMS'
        try:
            with open(writefile, 'w') as wf:
                wf.write(output)
        except Exception:
            logger.error("Failed to write "+wfstr+" file")
            raise
        logger.debug("Wrote to "+wfstr+" successfully.")
    return numlist, blocks, totalbeams


def writeAUXEXPBEAMS(beams, filename="AUXEXPBEAMS", header="Unknown system",
                     write=True, numbers=False, version=0, output_format='12E14.5'):
    """Takes a list of Beam objects and writes them in the format required by
    TensErLEED for experimental beam data. Returns the whole output as a
    string. 'numbers' defines whether a sequence of beam numbers in line 2
    is expected. 'version' is to pass the TensErLEED-version, only needed for
    new formatting styles."""
    output = header+"\n"
    if numbers:
        if version < 1.7:
            bformatter = ff.FortranRecordWriter("25I3")
        else:
            bformatter = ff.FortranRecordWriter("25I4")
        output += bformatter.write([n+1 for n in range(0, len(beams))]) + "\n"
        if len(beams) % 25 == 0:
            output += "\n"
    # Standard format used to be 12F6.2, but this may 
    # have too little dynamic range for theory-theory comparisons.
    # Instead we can use 12E14.5 which is the same as used everywhere else.
    if output_format not in ("12F6.2", "12E14.5"):
        logger.error(f"Incompatible format requested for AUXEXPEBAMS: {output_format}\n"
                     f"Possible are '12F6.2' and '12E14.5'.")
        raise RuntimeError
    
    output += f" ({output_format})\n"
    decimal_writer = ff.FortranRecordWriter(output_format)
    integer_writer = ff.FortranRecordWriter('I4')
    for beam in beams:
        # renormalize
        minintens = min(beam.intens.values())
        # Alex March 2022: This is quite important actually!!! Will mess up otherwise if negative values in EXPBEAMS
        offset = max(0, -minintens)  # if beams contain negative values, offset
        scaling = 999.99 / (max(beam.intens.values()) + offset)
        # working dict to hold normalized beams
        working_beam = {}
        for k in beam.intens:
            working_beam[k] = (beam.intens[k] + offset) * scaling
        # write
        output += "*"+beam.label.replace("|", " ") + "*\n"
        ol = integer_writer.write([len(working_beam)]).ljust(8)
        output += ol + '{:10.4E}\n'.format(scaling)
        # zip & flatten energies and values into one list
        outlist = [val for tup in working_beam.items() for val in tup]
        output += decimal_writer.write(outlist)
        if output[-1] != "\n":
            output += "\n"
    if write:
        try:
            with open(filename, 'w') as wf:
                wf.write(output)
        except Exception:
            logger.warning("Failed to write "+filename)
        logger.debug("Wrote to "+filename+" successfully")
    return output


def writeFdOut(beams, beamlist=None, filename="refcalc-fd.out",
               header="Unknown system"):
    """Writes an fd.out file for the given beams. Also returns the file "
    "contents as string."""
    out = header+"\n"
    # read BEAMLIST
    if beamlist is None:
        logger.warning("writeFdOut: no beamlist passed, attempting to read "
                       "BEAMLIST directly.")
        try:
            beamlist = readBEAMLIST("BEAMLIST")
        except Exception:
            logger.error("Error getting beamlist: file not found.")
            raise
    out += "{:>3}\n".format(len(beams))
    energies = set()
    for b in beams:
        for (i, line) in enumerate(beamlist[1:]):
            llist = line.split()
            if (len(llist) > 1
                    and b.isEqual_hk((float(llist[0]), float(llist[1])))):
                out += "{:>4}".format(i+1) + line[:23] + "\n"
                break
        else:
            logger.error("writeFdOut: passed beams contain beam {}, which "
                         "was not found in BEAMLIST.".format(b.label))
            raise ValueError("writeFdOut: beam {} is not in BEAMLIST."
                             .format(b.label))
        energies.update(b.intens.keys())
    for en in sorted(list(energies)):
        o = "{:7.2f} 0.0001".format(en)
        line_len = 1
        for b in beams:
            if en in b.intens:
                o += "{:14.5E}".format(b.intens[en])
            else:
                o += "{:14.5E}".format(0)
            line_len += 1
            if line_len == 5:
                o += "\n"
                line_len = 0
        out += o+"\n"
    try:
        with open(filename, "w") as wf:
            wf.write(out)
    except Exception:
        logger.warning("writeFdOut: Failed to write output to file {}."
                       .format(filename))
        raise
    return out



"""A collection of functions that run ViPErLEED from a provided ASE object."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-06-04'
__license__ = 'GPLv3+'

from collections import defaultdict
from dataclasses import FrozenInstanceError, dataclass
from io import StringIO
import logging
from numbers import Real
import os
from pathlib import Path
import shutil
from typing import Any, Sequence
import warnings

import numpy as np

from viperleed.calc import DEFAULT_WORK
from viperleed.calc.classes.atom_containers import AtomList
from viperleed.calc.classes.rparams import IVShiftRange
from viperleed.calc.classes.rparams import Rparams
from viperleed.calc.classes.rparams import TheoEnergies
from viperleed.calc.classes.slab import Slab
from viperleed.calc.files import iorfactor as rf_io
from viperleed.calc.files import parameters, poscar
from viperleed.calc.files.beams import readOUTBEAMS
from viperleed.calc.files.ivplot import plot_iv
from viperleed.calc.lib.base import rotation_matrix
from viperleed.calc.run import run_calc
from viperleed.extensions.error_codes import check_ierr

try:
    from viperleed.extensions.rfactor import r_factor_new as rf          # TODO: pylint complains if not compiled
except ImportError:
    _HAS_NEW_RFACTOR = False
else:
    _HAS_NEW_RFACTOR = True


_LOGGER = logging.getLogger(__name__)
_INPUT_FILES = (
    "PARAMETERS",
    "VIBROCC",
    "IVBEAMS",
    "PHASESHIFTS",
    "DISPLACEMENTS",
    "EXPBEAMS",
    "EXPBEAMS.csv",
    )


@dataclass
class SlabTransform:
    """A container for unit-cell transformations.

    Attributes
    ----------
    orthogonal_matrix : numpy.ndarray, optional
        Shape (3, 3). Can contain an arbitrary orthogonal transform
        (i.e., a combination of mirroring and rotations) that will
        be applied to the cell. The transformation is applied BEFORE
        scaling, i.e., using the original unit cell. The matrix (O)
        is applied to both the unit cell and all Cartesian atomic
        positions. The unit cell U (unit vectors as columns) is
        transformed to O @ U. Atom Cartesian coordinates (v) are
        transformed to O @ v. This attribute can be used, e.g., to
        apply a rotation or a mirroring of the slab.
    uc_scaling : float or Sequence of float, optional
        Scaling to be applied to the cell. Scaling is applied AFTER
        the transformation, i.e., along the new unit cell vectors. This
        can be used to apply an isotropic scaling in all directions or
        to stretch/compress along unit cell vectors in order to change
        lattice constants in some direction. If the scaling is given as
        a number or a 1-element sequence, an isotropic scaling will be
        applied to the unit cell and atom positions. If a sequence with
        three entries is provided, the scaling will be applied along
        the unit cell vectors in the given order. Specifically, given
        uc_scaling == (s1, s2, s3), the new unit cell vectors will be
        a' = a * s1, b' = b * s2, c' = c * s3. Default is None, i.e.,
        no scaling.
    cut_cell_c_fraction : float, optional
        Fractional position along the c vector to cut the slab.
        (ViPErLEED expects the cell to have the surface at the
        "top" -- i.e., at high z coordinates, and bulk-like layers
        at the bottom). Cutting will be performed along the vector
        c AFTER all transformations (see uc_transformation_matrix)
        are applied. Only the part of the cell above (i.e., >=)
        this value will be kept. Everything else is discarded.
        Default is zero.
    """

    orthogonal_matrix: Any = None
    uc_scaling: Any = None
    cut_cell_c_fraction: float = 0.


# We would really like to use frozen=True here, but since python 3.8.10
# a frozen cannot inherit from a non-frozen. We're forced to emulate
# the frozen-ness by overriding __setattr__ and __delattr__, and using
# the __init__ implementation for frozen dataclasses. It is NOT A GOOD
# IDEA TO INHERIT FROM THIS WORKAROUND CLASS.
@dataclass
class _ImmutableSlabTransform(SlabTransform):
    """A container for immutable slab transformations.

    This class is meant exclusively as the type for Slab
    transformations that cannot be represented via a left-
    multiplication of the unit cell with an orthogonal
    matrix. These transforms cannot perform any scaling
    nor cutting of the unit cell. They should be combined
    with SlabTransform instances for that purpose.
    """

    def __init__(self):
        """Initialize instance attributes."""
        object.__setattr__(self, 'orthogonal_matrix', None)
        object.__setattr__(self, 'uc_scaling', None)
        object.__setattr__(self, 'cut_cell_c_fraction', 0.)

    def __setattr__(self, name, value):
        """Raise FrozenInstanceError."""
        if name == '__doc__':
            # One exception to allow nicer 'help' for instances
            super().__setattr__(name, value)
            return
        raise FrozenInstanceError(f"cannot assign to field {name!r}")

    def __delattr__(self, name):
        """Raise FrozenInstanceError."""
        raise FrozenInstanceError(f"cannot assign to field {name!r}")


def run_from_ase(exec_path, ase_object, inputs_path=None,
                 slab_transforms=(SlabTransform(),), cleanup_work=False):
    """Perform a ViPErLEED calculation starting from an ase.Atoms object.

    This function is a core part of the API for ViPErLEED. It allows
    to run a calculation from an ASE (Atomic Simulation Environment)
    Atoms object. The ase.Atoms object, which is expected to contain
    a simulation cell, will be cut at a specified height fraction,
    transformed into a ViPErLEED Slab and written to a POSCAR file. The
    ViPErLEED calculation is then executed on the structure using the
    input files found either in `inputs_path` or `exec_path`. The type
    of calculation performed depends on the value of the RUN parameter
    in the PARAMETERS file. If RUN = 1, this yields reference electron
    scattering amplitudes and intensities. These are collected in CSV
    files, which are returned as strings for further processing.

    Parameters
    ----------
    exec_path: string or Path
        Path where to execute the ViPErLEED calculation. The directory
        needs to exist. Input files will be copied from `inputs_path`,
        if provided. Otherwise it is assumed that all input files are
        already in `exec_path`. Temporary files will be stored in a
        subdirectory called work.
    ase_object : ase.Atoms
        ASE-type object that contains the atoms for the slab to be used
        with ViPErLEED. This is expected to contain a simulation cell,
        which will be transformed using `slab_transforms`.
    inputs_path: str or Path or None, optional
        Path to directory containing input files (PARAMETERS, VIBROCC,
        etc.) for ViPErLEED calculation. If provided, files will be
        copied from `inputs_path` to `exec_path` (silently overwriting
        existing files there!). If None or not given, no copying
        occurs. Default is None.
    slab_transforms : SlabTransform or Sequence thereof, optional
        Sequence of transformation operations to be applied to the Slab
        constructed from ase_object. These include optional "change of
        basis" operations, optional `uc_scaling` for scaling the unit
        vector lengths, and a `cut_cell_c_fraction` for selecting only
        a portion of the `ase_object`. In addition, some special
        SlabTransform objects that swap unit cell axes can also be used
        (swap_a_b, swap_b_c, and swap_c_a). See help(SlabTransform) and
        help(swap_*_*) for further details. The transforms are applied
        in the order given. The `cut_cell_c_fraction` is taken only
        from the LAST SlabTransform given. Make sure that vectors a & b
        do not have any components in z after transformation as this is
        not allowed and will raise a ValueError. Default is a single
        SlabTransform(), corresponding to no transformation and no cut.
    cleanup_work : bool, optional
        Whether the work directory created during execution of the
        calculation is to be removed at the end. Default is False.

    Returns
    -------
    theobeams_file_str : str
        Contents of the file "THEOBEAMS.csv", i.e. the scattering
        intensities as calculated in the reference calculation. An
        empty string is returned if no reference calculation was run.
    amp_real_file_str : str
        Contents of the file "Complex_amplitudes_real.csv", i.e.,
        the real part of the scattering amplitudes as calculated
        in the reference calculation. An empty string is returned
        if no reference calculation was run or if the file was not
        found.
    amp_imag_file_str : str
        Contents of the file "Complex_amplitudes_imag.csv", i.e., the
        imaginary part of the scattering amplitudes as calculated in
        the reference calculation. An empty string is returned if no
        reference calculation was run or if the file was not found.
    v0i : float
        Imaginary part of the inner potential as read from the
        PARAMETERS file. This is returned here since it is an input
        parameter for the R-factor calculation. The value is read
        from PARAMETERS and left unchanged - it is not a result of
        the calculation and is returned here for convenience only.

    Raises
    ------
    FileNotFoundError
        If `exec_path` is non-existent or not a directory.
    FileNotFoundError
        If no PARAMETERS file was found in `exec_path`, after
        potentially copying over from `inputs_path`
    ValueError
        If either of the first two unit cell vectors have a
        component perpendicular to the surface (i.e., along
        the z coordinate) after transformation of the unit cell
    RuntimeError
        If the ViPErLEED calculation fails.

    Notes
    -------
    The parameter slab_transform and its `orthogonal_matrix` attribute
    can be used to apply some useful transformations to the Slab.
    Examples include rotations around an axis and flipping the cell
    along c (in case the lower part has to be kept, rather than the
    upper one). A few special SlabTransform objects are also provided
    for swapping unit-cell vectors. Have a look at the matrices
    provided as part of this module (use as the `orthogonal_matrix`
    attribute of a SlabTransform):
    -  rot_mat_x(theta) : Rotation matrix around x by theta (deg)
    -  rot_mat_z(theta) : Rotation matrix around z by theta (deg)
    -  rot_mat_axis(axis, theta): Rotation around axis by theta (deg)
    -  flip_c_mat : Mirror matrix that flips the cell along c
    and the special SlabTransform objects:
    -  swap_a_b : Swap vectors a & b (changes handedness)
    -  swap_b_c : Swap vectors b & c (changes handedness)
    -  swap_c_a : Swap vectors c & a (changes handedness)
    For applying multiple operations, the order does matter. You
    can: (i) Combine multiple orthogonal transformation matrices
    via matrix multiplication (@ symbol or np.dot) into a single
    SlabTransform object (into its `orthogonal_matrix` attribute):
    in this case the leftmost matrix is applied last. (ii) Provide
    a sequence of SlabTransform objects in `slab_transforms`: here
    transformations are applied in the order given (i.e., the
    leftmost is the applied first). (iii) A combination of the
    previous two.
    """
    exec_path = Path(exec_path).resolve()
    if not exec_path.is_dir():
        raise FileNotFoundError(
            f"Invalid exec_path={exec_path}: not "
            + "existent" if not exec_path.exists() else "a directory"
            )

    _copy_inputs_to_exec_path(inputs_path, exec_path)

    # Check for PARAMETERS file - without that we can't proceed
    # Files PHASESHIFTS and VIBROCC are not required and can be
    # generated if PARAMETERS are set correctly. IVBEAMS can be
    # generated if EXPBEAMS exists.
    parameters_file = exec_path / "PARAMETERS"
    if not parameters_file.is_file():
        # No PARAMETERS file – Error
        raise FileNotFoundError("No PARAMETERS file found – this is required")

    # Transfer ASE object into slab object for ViPErLEED,
    # and save it as a POSCAR file for debug purposes.
    slab = _make_and_check_slab(ase_object, slab_transforms)
    _write_poscar(slab, exec_path)

    # Take care of input files and work directory
    # NOTE: PARAMETERS should NOT contain SITE_DEF flags.
    # VIBROCC should contain *_surf sites for all elements.
    work_path = _make_work_dir(exec_path)

    home = Path.cwd()
    os.chdir(work_path)

    # Get temporary parameters object
    rparams = parameters.read(parameters_file)
    parameters.interpret(rparams, slab=slab, silent=False)

    # We are ready to run ViPErLEED! Have fun!
    try:
        exit_code = run_calc(slab=slab,
                             preset_params=_make_preset_params(rparams, slab))
    except Exception as err:
        # If ViPErLEED fails, move back to home directory
        os.chdir(home)
        raise RuntimeError("ViPErLEED calculation failed") from err

    if exit_code:
        os.chdir(home)
        raise RuntimeError("ViPErLEED calculation failed. "
                           "See log file for details.")

    # ViPErLEED should have succeeded if you arrive here. However, we may not
    # have run a refcalc (id == 1). In that case, return empty strings.
    if 1 not in rparams.RUN:
        os.chdir(home)
        return "", "", "", rparams.V0_IMAG

    # read out the THEOBEAMS.csv file and complex amplitudes
    try:
        content_list = _read_refcalc_output(rparams)
    finally:
        # Always move back home, even if there were errors
        os.chdir(home)

    if cleanup_work:
        try:
            shutil.rmtree(work_path)
        except OSError as err:
            _LOGGER.warning(
                f"Failed to remove work directory {work_path}. Info: {err}"
                )
    return (*content_list, rparams.V0_IMAG)


def _copy_inputs_to_exec_path(inputs_path, exec_path):
    """Copy all calc input files from inputs_path to exec_path."""
    if inputs_path is None:
        return
    inputs_path = Path(inputs_path)
    for file in _INPUT_FILES:
        try:
            shutil.copy2(inputs_path / file, exec_path / file)
        except FileNotFoundError:
            pass


def _apply_transform(slab, transform, apply_cut=False):
    """Apply a single SlabTransform `transform` to `slab`."""
    indices = None       # Unit-cell and position indices to be swapped
    if transform is swap_a_b:
        indices = (0, 1), (1, 0)
    elif transform is swap_b_c:
        indices = (2, 1), (1, 2)
    elif transform is swap_c_a:
        indices = (0, 2), (2, 0)

    if indices:
        # Swapping axes will require to make a new bulk slab for sure
        if not slab.is_bulk:
            slab.bulkslab = None
        new_axes, old_axes = ((ind,) for ind in indices)
        slab.ucell.T[new_axes] = slab.ucell.T[old_axes]
        for atom in slab:
            atom.pos[new_axes] = atom.pos[old_axes]
        return  # No scaling nor cutting

    if transform.orthogonal_matrix is not None:
        slab.apply_matrix_transformation(transform.orthogonal_matrix)
    if transform.uc_scaling is not None:
        if isinstance(transform.uc_scaling, Real):
            transform.uc_scaling = (transform.uc_scaling,)
        slab.apply_scaling(*transform.uc_scaling)

    if not apply_cut or not transform.cut_cell_c_fraction:
        return

    # Cut the slab as given in cut_cell_c_fraction - very important
    if transform.cut_cell_c_fraction:
        slab.atlist = AtomList(at for at in slab
                               if at.pos[2] >= transform.cut_cell_c_fraction)
    slab.update_atom_numbers()
    slab.update_element_count()


def _make_and_check_slab(ase_object, transforms):
    """Return a Slab from ase.Atoms with transformations applied."""
    slab = Slab.from_ase(ase_object)

    # Transformations of slab:
    # Mirror/rotation/swapping and/or stretching/shrinking
    if isinstance(transforms, SlabTransform):
        transforms = (transforms,)
    last_transform = transforms[-1]
    for transform in transforms:
        _apply_transform(slab, transform,
                         apply_cut=transform is last_transform)

    # Check if the now transformed slab has any z components in vectors
    # a & b. Error out as this would mess up parts of the TensErLEED
    # calculations, as we always pass on only the first two components.
    ab_vecs = slab.ucell.T[:2]
    if any(ab_vecs[:, 2] > 1e-5):
        raise ValueError(
            f"z component found in unit cell vector a ({ab_vecs[0]}) or b "
            f"({ab_vecs[1]}). This is not allowed in a TensErLEED calculation."
            " Check eventual transformations applied to the unit cell and make"
            " sure a and b are in the (x,y) plane."
            )
    return slab


def _make_preset_params(rparams, slab):
    """Return preset parameters to use when running viperleed.calc."""
    if rparams.SITE_DEF:
        return {}

    # When no sites are explicitly defined, give all
    # atoms visible from vacuum a '_surf' site label
    _LOGGER.warning("Parameter SITE_DEF not specified. Double check "
                    "PARAMETERS and POSCAR to make sure all sites are "
                    "recognized correctly. Default *_surf sites definitions "
                    "will be added for atoms visible from vacuum.")
    preset_params = {}
    site_def = defaultdict(list)
    for atom in slab.getSurfaceAtoms(rparams):
        site_def[atom.el].append(atom.num)
    preset_params["SITE_DEF"] = {
        element: {"surf": atom_numbers}
        for element, atom_numbers in site_def.items()
        }

    return preset_params


def _make_work_dir(exec_path):
    """Return exec_path/'work' after copying there all input files."""
    work_path = exec_path / DEFAULT_WORK
    work_path.mkdir(exist_ok=True)
    # copy input files to work directory
    for file in _INPUT_FILES:
        try:
            shutil.copy2(exec_path / file, work_path / file)
        except FileNotFoundError:
            pass
    return work_path


def _read_refcalc_output(rpars):
    """Return the contents of the output files produced by a refcalc."""
    # List of file name and earliest version in which it appeared
    output_files = (
        ("THEOBEAMS.csv", 0),
        ("Complex_amplitudes_real.csv", 1.73),
        ("Complex_amplitudes_imag.csv", 1.73),
        )

    content_list = []
    for filename, min_version in output_files:
        _path = Path(filename).resolve()
        content_str = ""
        if _path.is_file():
            with _path.open("r", encoding="utf-8") as fproxy:
                content_str = fproxy.read()
        elif rpars.TL_VERSION >= min_version:
            _LOGGER.error(f"Could not find file {filename}")
        content_list.append(content_str)
    return content_list


def _write_poscar(slab, exec_path):                                             # TODO: could be done with a flag on poscar.write
    """Save slab as a POSCAR file, but warn if one is already there."""
    poscar_path = exec_path / "POSCAR"
    if poscar_path.exists():
        _LOGGER.warning("A 'POSCAR' file is already present in "
                        f"{exec_path} and will be overwritten.")
    poscar.write(slab, poscar_path)


def rfactor_from_csv(                                                           ## TODO: add kwarg for mapping for averaging
    beams_files,
    v0i,
    beams_file_is_content=(False, False),
    v0r_shift_range=(-3, 3),
    intpol_deg=5,
    intpol_step=0.5,
    return_beam_arrays = False,
):
    """Compute the Pendry R-factor between two CSV files (in ViPErLEED format).

    Read in two CSV files containing LEED-I(V) spectra and compute the mutual
    R-factor. The format of the files must be the same as accepted/generated by
    ViPErLEED.

    Parameters
    ----------
    beams_files : tuple of two string
        The two CSV files to be read in for R-factor calculation.
        These can contain either theoretical or experimental beams.
        If beams_file_is_content is set as True, the string is expected
        to contain the contents of the CSV file, otherwise it should be
        a path to the file.
    v0i : float
        Imaginary part of the inner potential (in eV).
    beams_file_is_content : tuple of two bool, default=(False, False)
        If either value is set to True, it is assumed that the
        corresponding element of beams_file contains a string with
        the CSV contents rather than a path to the file. For reading,
        a StringIO will be used internally.
    v0r_shift_range : Sequence of two float, default=(-3, 3)
        During R-factor calculation, the two beam sets can be shifted
        in energy relative to one another. v0r_shift_range determines
        the minimum and maximum shifts (given in eV). The values
        should be integer multiples of intpol_step. If they are not,
        v0r_shift_range is expanded to the next integer multiples.
    intpol_deg : int, default=5
        Either 3 or 5; degree of the interpolating spline polynomial
    intpol_step : float, default=0.5
        Step size of the energy grid the beam data will be interpolated
        on (in eV).
    return_beam_arrays : bool, default = False
        For debugging, you can optionally return the interpolated beams,
        y-functions, number of overlapping points & per-beam R-factors.

    Returns
    -------
    best_R : float
        The R-factor between the two files. Returns the best R factor
        found after variation in v0r_shift_range.
    best_v0r : float
        The v0r value for which the best R-factor was found.
    """
    if not _HAS_NEW_RFACTOR:
        raise ModuleNotFoundError(
            "Missing R-factor compiled Fortran extension module. "
            "Run make in viperleed/extensions, then try again",
            name='viperleed.extensions.rfactor'
            )

    # Use Pendry R-factor - TODO: discuss if this should be a user
    # parameter in the API (if so, the code below will need a few
    # adjustments)
    which_r = 1

    # Check compilation of wrapped R-factor code
    check_ierr(rf.test_compilation())

    err_msg = ""

    # Check if file is supposed to be read from disk or from memory and pass
    # it to readOUTBEMS accordingly.
    beams_tmp = [None, None]
    for i, (is_content, file) in enumerate(zip(beams_file_is_content, beams_files)):
        if is_content:
            with StringIO(file) as fproxy:
                beams_tmp[i] = readOUTBEAMS(filename=fproxy, sep=",")
        else:
            beams_tmp[i] = readOUTBEAMS(filename=file, sep=",")

    beams1, beams2 = beams_tmp

    # MRiva: The following checks are potentially time consuming since they
    # require looping through beams. Perhaps we can add an extra keyword
    # argument that skips this part altogether. Then, of course, we have
    # to assume that the beam files contain exactly the same beams in
    # exactly the same order. Not sure the order is consistent between
    # consecutive runs, though (it will be if the order of the output is
    # the same as in IVBEAMS).
    # AMI: Could do, but compared to the refcalc required for the
    # theoretical specetra, it's still very fast.

    # Check if multiple beams have the same index. This should never be
    # the case and would throw off the calculation.

    for i, beams in enumerate((beams1, beams2)):
        unique_lables = set()
        for beam in beams:
            unique_lables.add(beam.hkfrac)
        if len(unique_lables) != len(beams):
            raise RuntimeError(
                f"Multiple beams sharing the same index encountered"
                f" in input beamset {i}!"
            )

    # Check for common beams
    beam_corr_1_2 = {}

    for i, beam1 in enumerate(beams1):
        for j, beam2 in enumerate(beams2):
            if beam1.hkfrac == beam2.hkfrac:
                beam_corr_1_2[beam1] = beam2

    for beam in beams1:
        if beam not in list(beam_corr_1_2.keys()):
            err_msg += f"Beam {beam.label} from file 1 has no correspondence.\n"
    for beam in beams2:
        if beam not in list(beam_corr_1_2.values()):
            err_msg += f"Beam {beam.label} from file 2 has no correspondence.\n"

    corr_beams1 = []
    corr_beams2 = []

    # Warn but don't raise
    if err_msg:
        warnings.warn(err_msg, category=RuntimeWarning)

    # Done with checking beam indices
    n_beams = len(beam_corr_1_2)

    for key, value in beam_corr_1_2.items():
        corr_beams1.append(key)
        corr_beams2.append(value)

    # Now get beams into arrays
    (beams1_en,
     beams1_id_start,
     beams1_n_E_beams,
     beams1_arr) = rf_io.beamlist_to_array(corr_beams1)

    (beams2_en,
     beams2_id_start,
     beams2_n_E_beams,
     beams2_arr) = rf_io.beamlist_to_array(corr_beams2)

    # Prepare an Rparams object to get the energy ranges right.
    # Treat corr_beams1 as 'experiment' and corr_beams2 as 'theory'
    rpars = Rparams()
    rpars.expbeams = corr_beams1
    rpars.THEO_ENERGIES = TheoEnergies.from_sorted_grid(beams2_en)
    rpars.IV_SHIFT_RANGE = IVShiftRange(*v0r_shift_range, intpol_step)

    # Finally, get the right energies
    _, theo_range, *_ = rf_io.prepare_rfactor_energy_ranges(
        rpars,
        n_expand=(intpol_deg - 1) // 2
        )
    grid = np.arange(theo_range.min,
                     theo_range.max + 0.1*intpol_step,
                     intpol_step)

    averaging_scheme = np.int32(np.arange(n_beams) + 1)  # don't average
    n_derivs = 1

    # interpolate and prepare beam arrays

    # initialize beam arrays
    (
        beams1_e_start_beams_out,
        beams1_n_e_beams_out,
        beams1_intpol_intensity,
        beams1_deriv_y,
        ierr,
    ) = rf.alloc_beams_arrays(n_beams, len(grid))
    check_ierr(ierr)

    ierr = rf.prepare_beams(
        beams1_en,
        beams1_arr,
        beams1_id_start + 1,
        beams1_n_E_beams,
        intpol_deg,
        n_derivs,
        grid,
        beams1_e_start_beams_out,
        beams1_n_e_beams_out,
        beams1_intpol_intensity,
        beams1_deriv_y,
    )
    check_ierr(ierr)

    beams1_der = beams1_deriv_y.copy()

    # writes y into deriv_y
    rf.pendry_y_beamset(
        beams1_intpol_intensity,
        beams1_deriv_y,
        beams1_e_start_beams_out,
        beams1_n_e_beams_out,
        v0i
    )

    # Beams 2
    (
        beams2_e_start_beams_out,
        beams2_n_e_beams_out,
        beams2_intpol_intensity,
        beams2_deriv_y,
        ierr,
    ) = rf.alloc_beams_arrays(n_beams, len(grid))
    check_ierr(ierr)

    ierr = rf.prepare_beams(
        beams2_en,
        beams2_arr,
        beams2_id_start + 1,
        beams2_n_E_beams,
        intpol_deg,
        n_derivs,
        grid,
        beams2_e_start_beams_out,
        beams2_n_e_beams_out,
        beams2_intpol_intensity,
        beams2_deriv_y,

    )
    check_ierr(ierr)

    beams2_der = beams2_deriv_y.copy()

    # writes y into deriv_y
    rf.pendry_y_beamset(
        beams2_intpol_intensity,
        beams2_deriv_y,
        beams2_e_start_beams_out,
        beams2_n_e_beams_out,
        v0i
    )

    # Ready to calculate R-factor
    *bounds, _ = rpars.IV_SHIFT_RANGE
    v0r_shift_range_int = np.array([round(b / intpol_step) for b in bounds],
                                   dtype=np.int32)
    v0r_center = sum(v0r_shift_range_int) // 2
    start_guess = np.array(
        [v0r_shift_range_int[0], v0r_center, v0r_shift_range_int[1]],
        dtype=np.int32
        )

    # Same settings as used in R-factor for ViPErLEED
    fast_search = False  # TODO change to True as default if we are sure it is better
    tol_r = (1 - 5e-2,)
    tol_r_2 = (1 - 5e-2,)
    max_fit_range = 6

    (_, best_v0r, best_r, n_evaluated, r_beams, numerators, denominators, n_overlapping_points, ierr) = rf.r_beamset_v0r_opt_on_grid(
        which_r,
        v0r_shift_range_int,
        start_guess,
        fast_search,
        intpol_step,
        beams1_deriv_y,
        beams2_deriv_y,
        beams1_e_start_beams_out,
        beams2_e_start_beams_out,
        beams1_n_e_beams_out,
        beams2_n_e_beams_out,
        tol_r=tol_r,
        tol_r_2=tol_r_2,
        max_fit_range=max_fit_range,
    )
    # If encountered an error
    check_ierr(ierr)

    #  for debugging, we have the option of returning the beam_arrays, y-Functions etc.
    if (return_beam_arrays):
        return best_r, best_v0r, (beams1_intpol_intensity, beams2_intpol_intensity), (beams1_deriv_y, beams2_deriv_y), r_beams, n_overlapping_points, (beams1, beams2)

    return best_r, best_v0r


def plot_iv_from_csv(
    beam_file,
    output_file=None,
    beam_file_is_content=False,
    which_beams=[],
    legends=None
):
    """Plots LEED-I(V) spectra directly from a CSV file.

    This function can be used to plot LEED-I(V) spectra directly from
    a CSV file without running any other calculations.
    It is assumed that the CSV file uses the format used by ViPErLEED.
    Returns nothing but writes the pdf file.


    Parameters
    ----------
    beam_file : string
        If beam_file_is_content is set as True, the string is expected
        to contain the contents of the CSV file, otherwise it should be
        a path to the file.
    beam_file_is_content : bool, default = False
        If set to True, it is assumed that the beam_file
        contains a string with the CSV contents rather than a path to
        the file. For reading, a StringIO will be used internally.
    output_pdf : pathlib Path, string or None; default = None
        Path to the PDF file where the plot will be saved, if Path or string.
        (File is created.) If None, returns list of matplotlib figures.
    which_beams : range or Sequence of int or strings
        Indices specifying which beams to plot. Order is the same as in the csv.
    legends : None or list of strings, default = None
        Legends to be used for the various files.

    Raises
    ------
    RuntimeError
        If which_beams is not Sequence containing just integers or
        strings.
    """
    if isinstance(beam_file, Sequence):
        n_beams = len(beam_file)
        assert len(beam_file_is_content) == len(beam_file)
    else:
        n_beams = 1
        assert type(beam_file_is_content) is bool

    all_beam_data = []
    labels = []
    for file, is_content in zip(beam_file, beam_file_is_content):
        tmp_file = StringIO(file) if is_content else file
        beam_list = readOUTBEAMS(filename=tmp_file, sep=",")
        if not beam_list:
            continue

        if which_beams == [] or which_beams == "all" or which_beams is None:
            pass
        elif isinstance(which_beams, range) or all(
            [type(label) is int for label in which_beams]
        ):
            beam_list = [beam_list[i] for i in which_beams]
        elif all([type(label) is str for label in which_beams]):
            new_beam_list = []
            for beam in beam_list:
                if beam.label in which_beams:
                    new_beam_list.append(beam)
            beam_list = new_beam_list
        else:
            raise RuntimeError("Invalid argument which_beams in plot_iv_from_csv()")

        # normalize each beam by the max intensity
        for beam in beam_list:
            maxint = max(beam.intens.values())
            beam.intens = {en: i / maxint for en, i in beam.intens.items()}
        beam_labels=[beam.label for beam in beam_list]
        all_beam_data.append(beam_list)
        if not labels:
            labels = beam_labels
        else:
            for jj in range(len(beam_labels)):
                labels[jj] += ", " + beam_labels[jj]

    if not all_beam_data:
        # Nothing to plot
        raise ValueError("No beams to plot in beam_file")

    return plot_iv(all_beam_data, output_file, labels=labels, legends=legends)


def rot_mat_x(theta):
    """Return a rotation matrix around the x axis.

    The rotation is positive, i.e., clockwise when looking along x,
    when applied from the left to column vectors. The same rotation
    for row vectors can be obtained by multiplying on the right with
    the transpose of the return value.

    Parameters
    ----------
    theta : float
        Angle of rotation in degrees.

    Returns
    -------
    rot_mat : numpy.ndarray
        Shape (3, 3). Rotation matrix around the x axis.
    """
    theta = np.radians(theta)
    rot_mat = np.identity(3)
    rot_mat[1:, 1:] = rotation_matrix(theta, dim=2)
    return rot_mat


def rot_mat_z(theta):
    """Return a rotation matrix around the z axis.

    The rotation is positive, i.e. clockwise when looking along z,
    when applied from the left to column vectors. The same rotation
    for row vectors can be obtained by multiplying on the right with
    the transpose of the return value.

    Parameters
    ----------
    theta : float
        Angle of rotation in degrees.

    Returns
    -------
    rot_mat : numpy.ndarray
        Shape (3, 3). Rotation matrix around the z axis.
    """
    theta = np.radians(theta)
    return rotation_matrix(theta, dim=3)


def rot_mat_axis(axis, theta):
    """Return a 3D rotation matrix by `theta` around `axis`.

    Parameters
    ----------
    axis : numpy.ndarray
        A vector parallel to the rotation axis. Shape (3,).
    theta : float
        Rotation angle in degrees.

    Returns
    -------
    rotation_matrix : numpy.ndarray
        Rotation matrix around `axis` by `theta`. The rotation is
        counter-clockwise when applied to column vectors from
        the left.
    """
    # We use the 'concise form' of the axis--angle formulation from
    # en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    axis_dir = axis / np.linalg.norm(axis)

    # Cross-product matrix from stackoverflow.com/questions/66707295
    u_cross = np.cross(axis_dir, -np.identity(3))
    u_outer = np.outer(axis_dir, axis_dir)
    theta = np.radians(theta)
    return (np.cos(theta) * np.identity(3)
            + np.sin(theta) * u_cross
            + (1 - np.cos(theta)) * u_outer)


# Some other useful transformations
# Swapping unit cell vectors
swap_a_b = swap_b_a = _ImmutableSlabTransform()
swap_b_c = swap_c_b = _ImmutableSlabTransform()
swap_c_a = swap_a_c = _ImmutableSlabTransform()


# Notice: the __doc__ edit will work fine in python >=3.10, but earlier
# versions print out type(obj).__doc__ when help is called. There's no
# good way around this, other than making each a separate class.
_FMT_DOC = """
    This SlabTransform swaps the '{}' and '{}' unit cell vectors,
    but does not change the Cartesian coordinates of the atoms.

    Be careful, as this transform WILL CHANGE THE HANDEDNESS of the
    unit cell. To avoid this, you can instead use a rotation around an
    appropriate axis. For orthogonal cells, you can use a 90 degrees
    rotation around '{}'.
"""

# pylint: disable=no-member
# Perhaps a bug? __doc__ exists and is set properly
swap_a_b.__doc__ += _FMT_DOC.format('a', 'b', 'c')
swap_b_c.__doc__ += _FMT_DOC.format('b', 'c', 'a')
swap_c_a.__doc__ += _FMT_DOC.format('c', 'a', 'b')
# pylint: enable=no-member


# flip the cell along c (useful if the lower half of the cell is to be kept)
flip_c_mat = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])



"""
======================================
  ViPErLEED Graphical User Interface
======================================
      *** module guilib.helpers ***

Created: 2021-03-11
Author: Michele Riva

This module collects helper functions used in various places in the GUI
"""


import numpy as np


def two_by_n_array_to_tuples(two_by_n, axis=None):
    """
    Converts a 2xN or Nx2 array into a zip object, that will return
    tuples when iterated over

    Parameters
    ----------
    two_by_n : numpy.ndarray
        Can have shape (N, 2) or (2, N). The output will always be an iterator
        of tuples, independent of the shape of the input
    axis : 0, 1, or None
        Which axis contains the 2-tuples. Use axis=0 if two_by_n[i] is the
        i-th tuple, axis=1 if two_by_n[:, i] is the i-th tuple. If None or
        omitted, the axis is inferred from which of the dimensions is equal
        to two, unless both are. In that case axis is mandatory.

    Returns
    -------
    zip-object
    """
    # Make sure we get an array. It would be possible to also do this with
    # other array-like objects but it's not implemented
    if not isinstance(two_by_n, np.ndarray):
        raise TypeError("Need a numpy array as input")
    if (all(length != 2 for length in two_by_n.shape)
        or len(two_by_n.shape) != 2):
        raise ValueError("Invalid shape of index array. Expected (2, N)"
                             f" or (N, 2), found {two_by_n.shape}")
    if axis is None:
        # determine which one is the right axis
        if two_by_n.shape[0] == 2 and two_by_n.shape[1] != 2:
            axis = 0
        elif two_by_n.shape[0] != 2 and two_by_n.shape[1] == 2:
            axis = 1
        else:
            raise RuntimeError("Cannot determine automatically which is the"
                               "correct axis, as two_by_n.shape is (2, 2). "
                               "Provide the correct axis with the axis "
                               "optional argument")
    if axis == 0:  # Shape == (2, N)
        return zip(two_by_n[0], two_by_n[1])
    # Shape == (N, 2)
    return zip(two_by_n[:, 0], two_by_n[:, 1])


def two_by_two_array_to_tuple(two_by_two):
    """
    Convenience function that converts a 2x2 array to a 2x2 tuple. Does no
    type or shape checking!
    """
    return tuple(map(tuple, two_by_two))


def conventional_angles(theta, phi):
    """
    Given two angles theta and phi in degrees, returns a modified version such
    that 0 <= theta <= 180 and 0 <= phi < 360, and such that they represent
    the same vector as before

    Parameters
    ----------
    theta : float
        polar angle in degrees, negative values affect the value of phi
    phi : float
        azimuthal angle in degrees, returned modulo 360

    Returns
    -------
    (theta, phi) : (float, float)
    """
    phi %= 360
    if theta < 0:
        # need to change the sign of theta, and adapt phi accordingly
        theta *= -1
        phi -= 180
        phi %= 360
    return theta, phi


def remove_duplicates(data, return_type=None):
    """
    Removes duplicates from the input. The function preserves order, but works
    with generator objects only if all the items are hashable.

    Parameters
    ----------
    data : iterable
    return_type : callable, optional (default: same as input or tuple)
        A callable that can instantiate an iterable, which is then returned.
        If not given, the same type as the input is returned. It is generally
        safer to pass return_type, as only a few cases are handled as expected.
        Falls back on returning a tuple in case return_type would raise abs
        TypeError. Does not work as expected for multi-dimensional numpy arrays!

    Returns
    -------
    uniques : return_type if given, type(data) if it doesn't raise errors,
              tuple otherwise
    """
    # try to use an order-preserving O(1) algorithm
    try:
        ret = dict.fromkeys(data)
    except TypeError:
        # fall back to an O(n^2) algorithm in case there are unhashable elements
        ret = [d for i, d in enumerate(data) if d not in data[:i]]

    if return_type is None:
        if isinstance(data, np.ndarray):
            return_type = np.array
        else:
            return_type = type(data)

    if return_type is str:
        return_type = lambda x: "".join(map(str, x))
    elif return_type in (np.array, np.asarray):
        return_type = lambda x: np.fromiter(x, data.dtype)

    try:
        return_type(ret)
    except (TypeError, ValueError):
        return_type = tuple
    return return_type(ret)


"""Package calc of viperleed.

This package contains all the main functionality of the TensErLEED
manager part of ViPErLEED, i.e., the part of ViPErLEED dedicated to
the calculation of theoretical I(V) curves as well as the optimization
of structural models such that their calculated I(V) curves best
fit experimental ones.

This package can be invoked as
>>> /path/to/python3 -m viperleed.calc <options>

Modules
-------
bookkeeper (script)
    Functionality for keeping track of
    calculation history, as in a logbook.
from_ase:
    API for executing viperleed.calc from an ase.Atoms object.
psgen
    Handle input/output to the FORTRAN program(s) that generate
    atomic phase-shifts and information concerning the energy
    dependence of the inner potential of the crystal.
run
    Main functionality for running calc from a set of input files.
symmetry
    Function for detecting symmetry of 2D- and 3D-periodic slabs.

Packages
--------
classes
    Definition of LEED-related classes used throughout viperleed.calc.
files
    Input/output handling for both TensErLEED sections and calc.
lib
    Functions and classes used in various parts of calc.
sections
    Functionality to run the various, logically different parts
    of TensErLEED.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-03'
__license__ = 'GPLv3+'

import logging

DEFAULT_HISTORY = 'history'
DEFAULT_WORK = 'work'
DEFAULT_WORK_HISTORY = 'workhistory'
LOGGER = logging.getLogger(__name__)
LOG_PREFIX = 'viperleed-calc'
ORIGINAL_INPUTS_DIR_NAME = 'original_inputs'
